import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.openeye.canonical.oemol import (
    CanonicalOemol,
)

representation_name = "canonical_oemol"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, CanonicalOemol)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_compute(fixture_representation):
    """That default scoring gives expected results."""
    representation = fixture_representation
    samples = ["C=C"]
    result = representation.featurise(samples)
    expected_result = [
        b"\x0b\xb5\n\x93\x19\x85mol_0\x13\x87\x82CC\x81\x00\x01\x02 \x81\x00\x10"
        b"\x9b\x81\x80?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00!\x81\x00",
    ]
    assert representation_name in result
    assert result[representation_name] == expected_result


@pytest.mark.parametrize(
    ["sample", "expected_output_sample", "tautomer_options"],
    [
        (
            "C/C=C(/O)F",
            b'\x0b\xe5\n\x9f\x19\x85mol_0\x13\x93\x85CCCOF\x84\x00\x01\x01\x01\x02"\x02\x03\x01\x02\x04\x01 \x81\x00\x10\xbf\x81\x80?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00!\x81\x00',
            {"save_stereo": True},
        ),
        (
            "C/C=C(/O)F",
            b"\x0bm\x81\n\xeb\x1f\xb6\x01\xb4\x01?\x00\x8d_stableScore_\x08\x00\x00\x00\x00\x00\x00\x00\xc0\x00\x8f_tautomerScore_\x08\x00\x00\x00\x00\x00@\x8f@\x19\x85mol_0\x13\xa7\x8aCCCOFHHHHH\x89\x00\x01\x01\x01\x02!\x02\x03\x02\x02\x04\x01\x00\x05\x01\x00\x06\x01\x00\x07\x01\x01\x08\x01\x01\t\x01 \x81\x00\x10\xfb\x81\x80?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00!\x81\x00",
            {},
        ),
    ],
)
def test_with_tautomer_options(
    fixture_representation,
    sample,
    expected_output_sample,
    tautomer_options,
):
    """That tautomer options are correctly applied"""
    representation = fixture_representation
    samples = [sample]
    expected_result = [expected_output_sample]

    result = representation.featurise(
        samples,
        reasonable_tautomer=True,
        tautomer_options=tautomer_options,
    )

    assert representation_name in result
    assert result[representation_name] == expected_result



from typing import Any

try:
    from rdkit.Chem.AllChem import GetMorganFingerprint

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
Raw/Unfolded Morgan fingerprint.

See the description of the `morgan` fingerprint for more information.
"""


class MorganUnfolded(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        radius: int = 3,
        invariants: list | None = None,
        from_atoms: list | None = None,
        use_chirality: bool = False,
        use_bond_types: bool = True,
        use_features: bool = False,
        use_counts: bool = False,
        bit_info: dict | None = None,
        include_redundant_environments: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[dict[str, int]]]:
        """Featurises the input molecules as unfolded Morgan fingerprints.

        Args:
            samples: The molecules to be fingerprinted.
            radius: The number of iterations to grow the fingerprint.
            invariants: The set of atom invariants to be used. Defaults to
                `None`, which corresponds to ECFP-type invariants.
            from_atoms: If provided, only the atoms in the vector will be used
                as centers in the fingerprint. Defaults to `None`.
            use_chirality: If set, additional information will be added to the
                fingerprint when chiral atoms are discovered, generating
                different fingerprints. Defaults to `False`.
            use_bond_types: If set, bond types will be included as part of the
                hash for calculating bits. Defaults to `True`.
            use_features: Defaults to `False`.
            use_counts: If set, counts of the features will be used.
            bit_info: Defaults to `None`.
            include_redundant_environments: If not None, the check for redundant
                atom environments will not be done. Defaults to `False`.

        Returns:
            Dict[str, List[Dict]]
                inputs featurised as unfolded (sparse) fingerprint (returned as a dictionary of non-zero molfluxs)

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation("morgan_unfolded")
            >>> samples = ["CCCC"]
            >>> representation.featurise(samples)
            {'morgan_unfolded': [{'1173125914': 1, '1244535424': 1, '2245384272': 1, '2246728737': 1, '3542456614': 1}]}

            >>> from molflux.features import load_representation
            >>> representation = load_representation("morgan_unfolded")
            >>> samples = ["CCCC"]
            >>> representation.featurise(samples, use_counts=True)
            {'morgan_unfolded': [{'1173125914': 2, '1244535424': 1, '2245384272': 2, '2246728737': 2, '3542456614': 2}]}

            >>> from molflux.features import load_representation
            >>> representation = load_representation("morgan_unfolded")
            >>> samples = ["CCCC"]
            >>> representation.featurise(samples, use_features=True)
            {'morgan_unfolded': [{'0': 1, '2602795547': 1, '3205495869': 1, '3766532888': 1}]}

            >>> from molflux.features import load_representation
            >>> representation = load_representation("morgan_unfolded")
            >>> samples = ["CCCC"]
            >>> representation.featurise(samples, use_features=True, use_counts=True)
            {'morgan_unfolded': [{'0': 4, '2602795547': 1, '3205495869': 2, '3766532888': 2}]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        unfolded_morgan_fp_list: list[dict[str, int]] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)
                rd_fp = GetMorganFingerprint(
                    mol,
                    radius=radius,
                    invariants=invariants or [],
                    fromAtoms=from_atoms or [],
                    useChirality=use_chirality,
                    useBondTypes=use_bond_types,
                    useFeatures=use_features,
                    useCounts=use_counts,
                    bitInfo=bit_info or {},
                    includeRedundantEnvironments=include_redundant_environments,
                )

                unfolded_morgan_fp_list.append(
                    {str(k): v for k, v in rd_fp.GetNonzeromolfluxs().items()},
                )

        return {self.tag: unfolded_morgan_fp_list}



from typing import Any, Literal

from pydantic.v1 import dataclasses

import datasets
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnModelBase

try:
    from sklearn.neighbors import KNeighborsRegressor
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn K nearest neighbours regressor.

Neighbors-based classification is a type of instance-based learning or non-generalizing
learning: it does not attempt to construct a general internal model, but simply stores
instances of the training data. Regression is computed from a simple majority vote
of the nearest neighbors of each point: a query point is assigned the average of the
nearest neighbors of the point.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_neighbors : int, default=5
    Number of neighbors to use by default for :meth:`kneighbors` queries.
weights : {'uniform', 'distance'} or callable, default='uniform'
    Weight function used in prediction.  Possible values:
    - 'uniform' : uniform weights.  All points in each neighborhood
      are weighted equally.
    - 'distance' : weight points by the inverse of their distance.
      in this case, closer neighbors of a query point will have a
      greater influence than neighbors which are further away.
    - [callable] : a user-defined function which accepts an
      array of distances, and returns an array of the same shape
      containing the weights.
    Uniform weights are used by default.
algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, default='auto'
    Algorithm used to compute the nearest neighbors:
    - 'ball_tree' will use :class:`BallTree`
    - 'kd_tree' will use :class:`KDTree`
    - 'brute' will use a brute-force search.
    - 'auto' will attempt to decide the most appropriate algorithm
      based on the values passed to :meth:`fit` method.
    Note: fitting on sparse input will override the setting of
    this parameter, using brute force.
leaf_size : int, default=30
    Leaf size passed to BallTree or KDTree.  This can affect the
    speed of the construction and query, as well as the memory
    required to store the tree.  The optimal value depends on the
    nature of the problem.
p : int, default=2
    Power parameter for the Minkowski metric. When p = 1, this is
    equivalent to using manhattan_distance (l1), and euclidean_distance
    (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.
metric : str or callable, default='minkowski'
    The distance metric to use for the tree.  The default metric is
    minkowski, and with p=2 is equivalent to the standard Euclidean
    metric. For a list of available metrics, see the documentation of
    :class:`~sklearn.metrics.DistanceMetric` and the metrics listed in
    `sklearn.metrics.pairwise.PAIRWISE_DISTANCE_FUNCTIONS`. Note that the
    "cosine" metric uses :func:`~sklearn.metrics.pairwise.cosine_distances`.
    If metric is "precomputed", X is assumed to be a distance matrix and
    must be square during fit. X may be a :term:`sparse graph`,
    in which case only "nonzero" molfluxs may be considered neighbors.
metric_params : dict, default=None
    Additional keyword arguments for the metric function.
n_jobs : int, default=None
    The number of parallel jobs to run for neighbors search.
    ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
    ``-1`` means using all processors.
    for more details.
    Doesn't affect :meth:`fit` method.
"""
Weights = Literal["uniform", "distance"]
Algorithm = Literal["auto", "ball_tree", "kd_tree", "brute"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class KNNRegressorConfig(ModelConfig):
    n_neighbors: int = 5
    weights: Weights = "distance"
    algorithm: Algorithm = "auto"
    leaf_size: int = 30
    p: int = 2
    metric: str = "minkowski"
    metric_params: dict | None = None
    n_jobs: int | None = None


class KNNRegressor(SKLearnModelBase[KNNRegressorConfig]):
    @property
    def _config_builder(self) -> type[KNNRegressorConfig]:
        return KNNRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> KNeighborsRegressor:
        config = self.model_config
        return KNeighborsRegressor(
            n_neighbors=config.n_neighbors,
            weights=config.weights,
            algorithm=config.algorithm,
            leaf_size=config.leaf_size,
            p=config.p,
            metric=config.metric,
            metric_params=config.metric_params,
            n_jobs=config.n_jobs,
        )

    def _train(
        self,
        train_data: datasets.Dataset,
        **kwargs: Any,
    ) -> Any:
        # perform validation of data
        n_data_points = len(train_data)
        n_neighbors = self.model_config.n_neighbors
        if n_data_points < n_neighbors:
            raise RuntimeError(
                f"The training data has {n_data_points} points, but the n_neighbors "
                f"parameter of the model is {n_neighbors}. Lower the n_neighbors value "
                f"in your model config or use more data!",
            )

        return super()._train(train_data=train_data, **kwargs)



import warnings

import pandas as pd
import pytest
from pydantic.v1 import ValidationError
from sklearn.exceptions import DataConversionWarning

import datasets
from molflux.modelzoo.catalogue import list_models
from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.load import load_from_dict, load_from_yaml, load_model
from molflux.modelzoo.models.sklearn.random_forest_regressor import (
    RandomForestRegressor,
)
from molflux.modelzoo.protocols import Estimator, Model

# Testing the SKLearn model base on a concrete implementation (random forest)
model_name = "random_forest_regressor"


@pytest.fixture(scope="function")
def fixture_model() -> Model:
    return load_model(model_name)


def test_in_catalogue():
    """That the model is registered in the catalogue."""
    catalogue = list_models()
    all_names = [name for names in catalogue.values() for name in names]
    assert model_name in all_names


def test_is_mapped_to_correct_class(fixture_model):
    """That the model name is mapped to the appropriate class."""
    model = fixture_model
    assert isinstance(model, RandomForestRegressor)


def test_implements_protocol(fixture_model):
    """That the model implements the public Estimator protocol."""
    model = fixture_model
    assert isinstance(model, Estimator)


def test_forwards_init_kwargs_to_builder():
    """That keyword arguments get forwarded to the model initialiser."""
    name = model_name
    model = load_model(name=name, tag="pytest-tag")
    assert model.tag == "pytest-tag"


def test_validates_init_kwargs():
    """That keyword arguments are validated.

    This is to improve user experience by failing fast if keyword arguments
    that do not match the expected types are provided on initialisation.
    """
    name = model_name

    with pytest.raises(ValidationError, match="1 validation error"):
        load_model(
            name=name,
            tag="pytest-tag",
            x_features=["good"],
            y_features="not-a-list",
        )


def test_from_dict_returns_model():
    """That loading from a dict returns a Model."""
    name = model_name
    config = {
        "name": name,
        "config": {},
    }
    model = load_from_dict(config)
    assert isinstance(model, Estimator)


def test_from_minimal_dict():
    """That can provide a config with only required fields."""
    name = model_name
    config = {
        "name": name,
    }
    assert load_from_dict(config)


def test_from_dict_forwards_config_to_model_builder():
    """That config keyword arguments get forwarded to the model initialiser."""
    name = model_name
    config = {
        "name": name,
        "config": {
            "tag": "pytest-tag",
        },
    }
    model = load_from_dict(config)
    assert model.tag == "pytest-tag"


def test_from_yaml_returns_collection_of_models(fixture_path_to_assets):
    path = fixture_path_to_assets / "core" / "config.yml"
    models = load_from_yaml(path=path)
    assert len(models) == 2
    assert "random_forest_regressor" in models
    assert "custom_classifier" in models
    assert all(isinstance(model, Estimator) for model in models.values())
    assert models["custom_classifier"].config.get("max_features") == "log2"


def test_attempt_training_on_featureless_model_raises():
    """That an exception is raised if training on a model that does not define
    features.

    References:

    """
    name = model_name
    model = load_model(name=name)
    data = datasets.Dataset.from_pandas(pd.DataFrame({"x1": [1, 2, 3]}))
    with pytest.raises(ValueError, match="features"):
        model.train(train_data=data)


def test_training_dataset_must_contain_model_x_features():
    """That the dataset used for training must contain the x features defined by
    the model.
    """
    name = model_name
    x_features = ["x1", "x2"]
    model = load_model(name=name, x_features=x_features)
    data = datasets.Dataset.from_pandas(pd.DataFrame({"x1": [1, 2, 3]}))
    with pytest.raises(ValueError, match="Data does not have required feature: 'x2'"):
        model.train(train_data=data)


def test_training_dataset_must_contain_model_y_features():
    """That the dataset used for training must contain the y features defined by
    the model.
    """
    name = model_name
    x_features = ["x1"]
    y_features = ["y1", "y2"]
    model = load_model(name=name, x_features=x_features, y_features=y_features)
    data = datasets.Dataset.from_pandas(
        pd.DataFrame({"x1": [1, 2, 3], "y1": [1, 2, 3]}),
    )
    with pytest.raises(ValueError, match="Data does not have required feature: 'y2'"):
        model.train(train_data=data)


def test_attempt_training_on_featureless_y_model_raises():
    """That an exception is raised if training on a model that does not define
    y_features.

    References:
    """
    name = model_name
    model = load_model(name=name, x_features=["x1"])
    data = datasets.Dataset.from_pandas(pd.DataFrame({"x1": [1, 2, 3]}))
    with pytest.raises(ValueError, match="features"):
        model.train(train_data=data)


def test_attempt_training_on_featureless_x_model_raises():
    """That an exception is raised if training on a model that does not define
    x_features.

    References:
    """
    name = model_name
    model = load_model(name=name, y_features=["y1"])
    data = datasets.Dataset.from_pandas(pd.DataFrame({"x1": [1, 2, 3]}))
    with pytest.raises(ValueError, match="feature"):
        model.train(train_data=data)


@pytest.mark.skip(reason="re-implement the logic for checking whether model is trained")
def test_cannot_predict_if_model_not_trained():
    """That the dataset used for predictions must contain the x features defined by
    the model.
    """
    name = model_name
    x_features = ["x1"]
    model = load_model(name=name, x_features=x_features)
    data = datasets.Dataset.from_pandas(pd.DataFrame({"x1": [1, 2, 3]}))
    with pytest.raises(NotTrainedError):
        model.predict(data=data)


train_df = pd.DataFrame({"x1": [1.5], "y1": [2]})
predict_df = pd.DataFrame({"x1": [3]})


@pytest.mark.parametrize(
    "train_data",
    [
        train_df,
        datasets.Dataset.from_pandas(train_df),
    ],
)
def test_train_on_single_sample(train_data):
    """That can train on a single sample.

    References:
    """
    name = model_name
    model = load_model(name=name, x_features=["x1"], y_features=["y1"])
    model.train(train_data=train_data)
    assert True


@pytest.mark.parametrize(
    "train_data, predict_data",
    [
        (train_df, predict_df),
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(predict_df),
        ),
    ],
)
def test_predict_on_single_sample(train_data, predict_data):
    """That can predict on a single sample.

    References:
    """
    name = model_name
    model = load_model(name=name, x_features=["x1"], y_features=["y1"])
    model.train(train_data=train_data)
    assert model.predict(data=predict_data)


def test_training_models_with_single_y_feature_does_not_raise_data_conversion_warning():
    """That can train a single-feature model without DataConversionWarnings.

    References:
    """
    name = model_name
    dataset = datasets.Dataset.from_dict({"x1": [1, 2, 3], "y1": [3, 2, 1]})
    model = load_model(name=name, x_features=["x1"], y_features=["y1"])
    with warnings.catch_warnings():
        warnings.simplefilter("error", DataConversionWarning)
        model.train(dataset)



from typing import Any

import torch

from molflux.modelzoo.models.lightning.mlp_regressor.mlp_config import MLPConfig
from molflux.modelzoo.models.lightning.module import (
    LightningModuleBase,
    SingleBatchStepOutput,
)


class MLPModule(LightningModuleBase):
    def __init__(self, model_config: MLPConfig) -> None:
        super().__init__(model_config=model_config)

        if model_config.num_layers == 1:
            self.module = torch.nn.Linear(
                in_features=model_config.input_dim,
                out_features=model_config.num_tasks,
            )
        else:
            module_list = []
            module_list.append(
                torch.nn.Linear(
                    in_features=model_config.input_dim,
                    out_features=model_config.hidden_dim,
                ),
            )
            module_list.append(torch.nn.SiLU())
            for _ in range(model_config.num_layers - 2):
                module_list.append(
                    torch.nn.Linear(
                        in_features=model_config.hidden_dim,
                        out_features=model_config.hidden_dim,
                    ),
                )
                module_list.append(torch.nn.SiLU())

            module_list.append(
                torch.nn.Linear(
                    in_features=model_config.hidden_dim,
                    out_features=model_config.num_tasks,
                ),
            )

            self.module = torch.nn.Sequential(*module_list)

        self.mse_loss = torch.nn.MSELoss()

    def forward(self, *args: Any, **kwargs: Any) -> Any:
        return self.module(*args, **kwargs)

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
        """fit method for doing a predict step"""
        del batch_idx, dataloader_idx

        x, _ = batch
        outputs = self(x)

        return outputs

    def _training_step_on_single_source_batch(
        self,
        single_source_batch: Any,
        source_name: str | None,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> SingleBatchStepOutput:
        """Implements the training step on one minibatch.

        Returns a dict with a loss key and a batch size."""
        del source_name, batch_idx, args, kwargs

        x, y = single_source_batch
        outputs = self(x)
        loss = self.mse_loss(input=outputs, target=y)
        mae = (outputs - y).abs().mean()

        batch_size = x.shape[0]

        return loss, {"mae": mae}, batch_size

    def _validation_step_on_single_source_batch(
        self,
        single_source_batch: Any,
        source_name: str | None,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> SingleBatchStepOutput:
        del source_name, batch_idx, args, kwargs

        x, y = single_source_batch
        outputs = self(x)
        loss = self.mse_loss(input=outputs, target=y)
        mae = (outputs - y).abs().mean()

        batch_size = x.shape[0]

        return loss, {"mae": mae}, batch_size



import datasets


def test_flatten_indices():
    # dataset with array and nested array with defined length
    dataset = datasets.Dataset.from_dict(
        {
            "array_nested": [[[0] * 3] * 7] * 2000,
            "array": [[0] * 3] * 2000,
        },
        features=datasets.Features(
            {
                "array_nested": datasets.Sequence(
                    datasets.Sequence(
                        datasets.Value(dtype="int64", id=None),
                        length=3,
                        id=None,
                    ),
                    length=-1,
                    id=None,
                ),
                "array": datasets.Sequence(
                    datasets.Value(dtype="int64", id=None),
                    length=3,
                    id=None,
                ),
            },
        ),
    )

    # a filter operation which does nothing (checks that all nested arrays have len < 8, they have len == 7)
    dataset_filtered = dataset.filter(lambda x: len(x["array"]) < 8)

    # step that fails because of Batched=True in map
    dataset_filtered.flatten_indices()



"""Out of sample R2"""

import logging
from typing import Any

import evaluate
from sklearn.metrics import mean_squared_error

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Out of sample R2.

Best possible score is 1.0 and it can be negative (because the model can be
arbitrarily worse). This metric overcomes some of the limitations of R2.
While R2 is comparing a perfect model and a constant model, when evaluated on
a test set it uses the test labels so in some sense has information leakage.
To overcome this the out of sample version of R2 uses the mean of training
labels rather than test labels. The mean of the training labels must be provided
as an additional argument, `dummy_predictions`. Note that this metric could
conceivably exaggerate the performance of a general model on a project
test set compared to a project model.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.
    dummy_predictions: Estimated value from a dummy model constant on the training data.

Returns:
    oos_r2: The out of sample R2 score.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("out_of_sample_r2")
    >>> predictions = [2.5, 0.0, 2, 8]
    >>> references = [3, -0.5, 2, 7]
    >>> dummy_prediction = 3.125
    >>> metric.compute(predictions=predictions, references=references, dummy_prediction=dummy_prediction)
    {'out_of_sample_r2': 0.949...}
"""

_CITATION = """\
https://towardsdatascience.com/whats-wrong-with-r-squared-and-how-to-fix-it-7362c5f26c53
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class OutOfSampleR2(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
            reference_urls=[
                "https://towardsdatascience.com/whats-wrong-with-r-squared-and-how-to-fix-it-7362c5f26c53",
            ],
        )

    def _score(  # type: ignore[override]
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        dummy_prediction: float,
        **kwargs: Any,
    ) -> MetricResult:
        """Compute out-of-sample R-squared."""
        mse_pred = mean_squared_error(references, predictions)
        mse_dummy = mean_squared_error(
            references,
            [dummy_prediction] * len(predictions),
        )
        mse_dummy = max(mse_dummy, 10**-16)
        oos_r2 = 1 - (mse_pred / mse_dummy)
        return {self.tag: oos_r2}



from molflux.datasets.load import load_from_yaml


def test_from_yaml_returns_metrics(fixture_path_to_assets):
    path = fixture_path_to_assets / "config.yml"
    datasets = load_from_yaml(path=path)
    assert len(datasets) == 1
    assert "dataset-0" in datasets
    assert isinstance(datasets, dict)



from typing import Any

try:
    from rdkit.Avalon.pyAvalonTools import GetAvalonFP

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
Avalon fingerprints from rdkit.
"""


class Avalon(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        n_bits: int = 512,
        is_query: bool = False,
        reset_vect: bool = False,
        bit_flags: int = 15761407,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates the Avalon fingerprint for each input molecule.

        Args:
            samples: The molecules to be fingerprinted.
            n_bits:
            is_query:
            reset_vect:
            bit_flags:

        Returns:
            Avalon fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('avalon')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples, n_bits=16)
            {'avalon': [[0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        avalon_fp_list: list[list] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)
                rd_fp = GetAvalonFP(
                    mol,
                    nBits=n_bits,
                    isQuery=is_query,
                    resetVect=reset_vect,
                    bitFlags=bit_flags,
                )
                avalon_fp_list.append(rd_fp.ToList())

        return {self.tag: avalon_fp_list}



from collections import deque
from collections.abc import Iterator
from math import ceil
from typing import Any

import numpy as np

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable
from molflux.splits.utils import partition

_DESCRIPTION = """
Deterministic linear cross-validator with rotation of returned indices.
Useful for extremely small datasets not served by other strategies.
Given n_splits, a rotation of window = len(dataset)/n_splits is calculated.
Indices array is rotated by window after each split.
"""


class LinearSplitWithRotation(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        *,
        n_splits: int = 1,
        train_fraction: float = 0.8,
        validation_fraction: float = 0.1,
        test_fraction: float = 0.1,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            y (optional): The target variable for supervised learning problems.
            groups (optional): Group labels for the samples used while splitting the dataset.
            n_splits (optional): The number of splits to generate. Defaults to 1.
            train_fraction (optional): The proportion of the dataset to include in the train split.
                Defaults to 0.8.
            validation_fraction: The proportion of the dataset to include in the validation split.
                Defaults to 0.1.
            test_fraction: The proportion of the dataset to include in the test split.
                Defaults to 0.1.
        """
        np.testing.assert_almost_equal(
            train_fraction + validation_fraction + test_fraction,
            1.0,
        )
        window = ceil(len(dataset) / n_splits)

        train_cutoff, validation_cutoff = partition(
            dataset,
            train_fraction,
            validation_fraction,
        )
        indices = np.array(range(len(dataset)))
        dequeued_indices = deque(indices)

        for _ in range(n_splits):
            train_indices, validation_indices, test_indices = np.split(
                list(dequeued_indices),
                [train_cutoff, validation_cutoff],
            )
            dequeued_indices.rotate(window)
            yield train_indices, validation_indices, test_indices



from molflux.splits.naming import camelcase_to_snakecase


def test_camelcase_to_snakecase():
    """That camelcase to snakecase conversion returns as expected."""
    assert camelcase_to_snakecase("ClassNameABC") == "class_name_abc"


def test_camelcase_to_snakecase_invariant():
    """That camelcase to snakecase conversion is invariant on snakecase input."""
    assert camelcase_to_snakecase("class_name_abc") == "class_name_abc"


def test_camelcase_to_snakecase_starts_with_capital():
    """Camelcase to snakecase conversion on chained capitalised letters."""
    assert camelcase_to_snakecase("ABCClassName") == "abc_class_name"



import re

_uppercase_uppercase_re = re.compile(r"([A-Z]+)([A-Z][a-z])")
_lowercase_uppercase_re = re.compile(r"([a-z\d])([A-Z])")


def camelcase_to_snakecase(name: str) -> str:
    """Convert CamelCase string to snake_case."""
    name = _uppercase_uppercase_re.sub(r"\1_\2", name)
    name = _lowercase_uppercase_re.sub(r"\1_\2", name)
    return name.lower()



from collections.abc import Iterable
from typing import Any

from molflux.splits.catalogue import get_splitting_strategy_cls
from molflux.splits.errors import DuplicateKeyError
from molflux.splits.parsers import Spec, dict_parser, yaml_parser
from molflux.splits.strategy import SplittingStrategy
from molflux.splits.typing import PathLike

SplittingStrategies = dict[str, SplittingStrategy]


def load_splitting_strategy(name: str, **init_kwargs: Any) -> SplittingStrategy:
    """Loads a splitting strategy from the catalogue."""

    # Fetch relevant splitting strategy cls from catalogue
    splitting_strategy_cls = get_splitting_strategy_cls(name)

    # Instantiate splitting strategy object
    strategy = splitting_strategy_cls(**init_kwargs)

    return strategy


def _load_from_spec(spec: Spec) -> SplittingStrategy:
    """Loads a splitting strategy from a validated Spec."""

    # Build strategy
    strategy = load_splitting_strategy(name=spec.name, **spec.config)

    # update state
    strategy.update_state(**spec.presets)

    return strategy


def load_from_dict(dictionary: dict[str, Any]) -> SplittingStrategy:
    """Loads a splitting strategy from a config dict."""

    # Validate dictionary
    spec = dict_parser(dictionary=dictionary)

    return _load_from_spec(spec=spec)


def load_from_dicts(dictionaries: Iterable[dict[str, Any]]) -> SplittingStrategies:
    """Loads splitting strategies from an iterable of dicts."""

    strategies = (load_from_dict(dictionary) for dictionary in dictionaries)
    collection = _make_strategy_collection(strategies)
    return collection


def load_from_yaml(path: PathLike) -> SplittingStrategies:
    """Loads splitting strategies from a yaml config file."""

    specs = yaml_parser(path=path)
    strategies = (_load_from_spec(spec=spec) for spec in specs)
    collection = _make_strategy_collection(strategies)
    return collection


def _make_strategy_collection(
    strategies: Iterable[SplittingStrategy],
) -> SplittingStrategies:
    """Collects SplittingStrategy-es into a SplittingStrategies collection."""
    collection: SplittingStrategies = {}
    for strategy in strategies:
        tag = strategy.tag
        if tag in collection:
            raise DuplicateKeyError(
                f"Splitting strategy with key {tag!r} has already been added. You can add a unique custom tag to one of the strategies to add it to the collection.",
            )
        collection[tag] = strategy
    return collection



"""
Tests for API features advertised as part of the Metrics protocol.
"""

from molflux.metrics.metric import Metrics


def test_has_init():
    """That can be initialised."""
    assert hasattr(Metrics, "__init__")
    assert callable(Metrics.__init__)


def test_has_getitem():
    """That can get items."""
    assert hasattr(Metrics, "__getitem__")
    assert callable(Metrics.__getitem__)


def test_has_iter():
    """That can be iterated over."""
    assert hasattr(Metrics, "__iter__")
    assert callable(Metrics.__iter__)


def test_has_setitem():
    """That can set items."""
    assert hasattr(Metrics, "__setitem__")
    assert callable(Metrics.__setitem__)


def test_has_add_batch():
    """That has an add_batch method."""
    assert hasattr(Metrics, "add_batch")
    assert callable(Metrics.add_batch)


def test_has_add_metric():
    """That has an add_metric method."""
    assert hasattr(Metrics, "add_metric")
    assert callable(Metrics.add_metric)


def test_has_compute():
    """That has a compute method."""
    assert hasattr(Metrics, "compute")
    assert callable(Metrics.compute)



from __future__ import annotations

import difflib
import itertools
from typing import TYPE_CHECKING, Any, Literal, get_args

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import (
    rdkit_mol_from_smiles,
    to_smiles,
)
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

if TYPE_CHECKING:
    from molflux.features.typing import MolArray

try:
    from rdkit.ML.Descriptors import MoleculeDescriptors

except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None


# from [descriptor[0] for descriptor in rdkit.Chem.Descriptors._descList]:
_Descriptor2D = Literal[
    "MaxAbsEStateIndex",
    "MaxEStateIndex",
    "MinAbsEStateIndex",
    "MinEStateIndex",
    "qed",
    "SPS",
    "MolWt",
    "HeavyAtomMolWt",
    "ExactMolWt",
    "NumValenceElectrons",
    "NumRadicalElectrons",
    "MaxPartialCharge",
    "MinPartialCharge",
    "MaxAbsPartialCharge",
    "MinAbsPartialCharge",
    "FpDensityMorgan1",
    "FpDensityMorgan2",
    "FpDensityMorgan3",
    "BCUT2D_MWHI",
    "BCUT2D_MWLOW",
    "BCUT2D_CHGHI",
    "BCUT2D_CHGLO",
    "BCUT2D_LOGPHI",
    "BCUT2D_LOGPLOW",
    "BCUT2D_MRHI",
    "BCUT2D_MRLOW",
    "AvgIpc",
    "BalabanJ",
    "BertzCT",
    "Chi0",
    "Chi0n",
    "Chi0v",
    "Chi1",
    "Chi1n",
    "Chi1v",
    "Chi2n",
    "Chi2v",
    "Chi3n",
    "Chi3v",
    "Chi4n",
    "Chi4v",
    "HallKierAlpha",
    "Ipc",
    "Kappa1",
    "Kappa2",
    "Kappa3",
    "LabuteASA",
    "PEOE_VSA1",
    "PEOE_VSA10",
    "PEOE_VSA11",
    "PEOE_VSA12",
    "PEOE_VSA13",
    "PEOE_VSA14",
    "PEOE_VSA2",
    "PEOE_VSA3",
    "PEOE_VSA4",
    "PEOE_VSA5",
    "PEOE_VSA6",
    "PEOE_VSA7",
    "PEOE_VSA8",
    "PEOE_VSA9",
    "SMR_VSA1",
    "SMR_VSA10",
    "SMR_VSA2",
    "SMR_VSA3",
    "SMR_VSA4",
    "SMR_VSA5",
    "SMR_VSA6",
    "SMR_VSA7",
    "SMR_VSA8",
    "SMR_VSA9",
    "SlogP_VSA1",
    "SlogP_VSA10",
    "SlogP_VSA11",
    "SlogP_VSA12",
    "SlogP_VSA2",
    "SlogP_VSA3",
    "SlogP_VSA4",
    "SlogP_VSA5",
    "SlogP_VSA6",
    "SlogP_VSA7",
    "SlogP_VSA8",
    "SlogP_VSA9",
    "TPSA",
    "EState_VSA1",
    "EState_VSA10",
    "EState_VSA11",
    "EState_VSA2",
    "EState_VSA3",
    "EState_VSA4",
    "EState_VSA5",
    "EState_VSA6",
    "EState_VSA7",
    "EState_VSA8",
    "EState_VSA9",
    "VSA_EState1",
    "VSA_EState10",
    "VSA_EState2",
    "VSA_EState3",
    "VSA_EState4",
    "VSA_EState5",
    "VSA_EState6",
    "VSA_EState7",
    "VSA_EState8",
    "VSA_EState9",
    "FractionCSP3",
    "HeavyAtomCount",
    "NHOHCount",
    "NOCount",
    "NumAliphaticCarbocycles",
    "NumAliphaticHeterocycles",
    "NumAliphaticRings",
    "NumAromaticCarbocycles",
    "NumAromaticHeterocycles",
    "NumAromaticRings",
    "NumHAcceptors",
    "NumHDonors",
    "NumHeteroatoms",
    "NumRotatableBonds",
    "NumSaturatedCarbocycles",
    "NumSaturatedHeterocycles",
    "NumSaturatedRings",
    "RingCount",
    "MolLogP",
    "MolMR",
    "fr_Al_COO",
    "fr_Al_OH",
    "fr_Al_OH_noTert",
    "fr_ArN",
    "fr_Ar_COO",
    "fr_Ar_N",
    "fr_Ar_NH",
    "fr_Ar_OH",
    "fr_COO",
    "fr_COO2",
    "fr_C_O",
    "fr_C_O_noCOO",
    "fr_C_S",
    "fr_HOCCN",
    "fr_Imine",
    "fr_NH0",
    "fr_NH1",
    "fr_NH2",
    "fr_N_O",
    "fr_Ndealkylation1",
    "fr_Ndealkylation2",
    "fr_Nhpyrrole",
    "fr_SH",
    "fr_aldehyde",
    "fr_alkyl_carbamate",
    "fr_alkyl_halide",
    "fr_allylic_oxid",
    "fr_amide",
    "fr_amidine",
    "fr_aniline",
    "fr_aryl_methyl",
    "fr_azide",
    "fr_azo",
    "fr_barbitur",
    "fr_benzene",
    "fr_benzodiazepine",
    "fr_bicyclic",
    "fr_diazo",
    "fr_dihydropyridine",
    "fr_epoxide",
    "fr_ester",
    "fr_ether",
    "fr_furan",
    "fr_guanido",
    "fr_halogen",
    "fr_hdrzine",
    "fr_hdrzone",
    "fr_imidazole",
    "fr_imide",
    "fr_isocyan",
    "fr_isothiocyan",
    "fr_ketone",
    "fr_ketone_Topliss",
    "fr_lactam",
    "fr_lactone",
    "fr_methoxy",
    "fr_morpholine",
    "fr_nitrile",
    "fr_nitro",
    "fr_nitro_arom",
    "fr_nitro_arom_nonortho",
    "fr_nitroso",
    "fr_oxazole",
    "fr_oxime",
    "fr_para_hydroxylation",
    "fr_phenol",
    "fr_phenol_noOrthoHbond",
    "fr_phos_acid",
    "fr_phos_ester",
    "fr_piperdine",
    "fr_piperzine",
    "fr_priamide",
    "fr_prisulfonamd",
    "fr_pyridine",
    "fr_quatN",
    "fr_sulfide",
    "fr_sulfonamd",
    "fr_sulfone",
    "fr_term_acetylene",
    "fr_tetrazole",
    "fr_thiazole",
    "fr_thiocyan",
    "fr_thiophene",
    "fr_unbrch_alkane",
    "fr_urea",
]
_ALL_AVAILABLE_DESCRIPTORS = list(get_args(_Descriptor2D))


_DESCRIPTION = f"""
209 rdkit 2d molecular descriptors.

The following descriptors are available: {_ALL_AVAILABLE_DESCRIPTORS!r}

For further info and references, see
https://www.rdkit.org/docs/GettingStartedInPython.html#list-of-available-descriptors.
"""


class RdkitDescriptors_2d(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        include: list[_Descriptor2D] | None = None,
        exclude: list[_Descriptor2D] | None = None,
        **kwargs: Any,
    ) -> dict[str, list[float]]:
        """Calculates 2D molecular descriptors.

        Args:
            samples: The molecules for which to calculate descriptors.
            include: A list of specific 2D descriptor names to calculate. If
                `None`, all available 2D descriptors are calculated. Defaults to
                `None`.
            exclude: A list of specific 2D descriptor names to not calculate. If
                `None`, no descriptors are excluded. Defaults to `None`.

        Returns:
            Calculated values of the 2D descriptor, each descriptor as its own
             feature.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('rdkit_descriptors_2d')
            >>> samples = ['COc1cc2c(cc1OCCCN3CCOCC3)c(ncn2)Sc4nccs4', 'c1ccccc1']
            >>> representation.featurise(samples, include=["SlogP_VSA8", "VSA_EState8"])
            {'rdkit_descriptors_2d::SlogP_VSA8': [10.90..., 0.0], 'rdkit_descriptors_2d::VSA_EState8': [5.25..., 0.0]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        # not using set difference to avoid possible internal shuffling
        descriptors_to_calculate = [
            x
            for x in (include or _ALL_AVAILABLE_DESCRIPTORS)
            if x not in (exclude or [])
        ]

        if not descriptors_to_calculate:
            raise ValueError(
                "No descriptors to calculate: please expand your 'select' filter and / or reduce your 'exclude' filter.",
            )

        invalid_descriptors = set(descriptors_to_calculate).difference(
            _ALL_AVAILABLE_DESCRIPTORS,
        )
        if invalid_descriptors:
            msg = "The following descriptor(s) are not available:"
            for invalid_descriptor in invalid_descriptors:
                msg += f"\n\t{invalid_descriptor!r}"
                similar = difflib.get_close_matches(
                    invalid_descriptor,
                    _ALL_AVAILABLE_DESCRIPTORS,
                )
                if similar:
                    msg += f" -> You might be looking for one of these: {similar}"
            raise ValueError(msg)

        calculator = MoleculeDescriptors.MolecularDescriptorCalculator(
            descriptors_to_calculate,
        )

        descriptors_results_list: list[list[float]] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smiles = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smiles)
                descriptors_results_list.append(calculator.CalcDescriptors(mol))

        # Transpose the results and return one feature for each descriptor
        return {  # type: ignore[var-annotated]
            f"{self.tag}::{name}": v
            for name, v in itertools.zip_longest(
                calculator.descriptorNames,
                map(list, zip(*descriptors_results_list, strict=False)),
                fillvalue=[],
            )
        }


def list_available_rdkit_descriptors_2d() -> list[str]:
    """Returns all available 2D descriptors names.

    This is a convenience function to abstract away the rdkit backend details.
    """
    return _ALL_AVAILABLE_DESCRIPTORS



import numpy as np
import pytest

from molflux.splits.catalogue import list_splitting_strategies
from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy

strategy_name = "linear_split"


@pytest.fixture(scope="module")
def fixture_test_strategy():
    return load_splitting_strategy(strategy_name)


@pytest.fixture(scope="module")
def fixture_sample_dataset():
    return np.random.rand(100)


def test_is_in_catalogue():
    """That the strategy is registered in the catalogue."""
    catalogue = list_splitting_strategies()
    all_strategy_names = [name for names in catalogue.values() for name in names]
    assert strategy_name in all_strategy_names


def test_implements_protocol(fixture_test_strategy):
    """That the strategy implements the protocol."""
    strategy = fixture_test_strategy
    assert isinstance(strategy, SplittingStrategy)


def test_yields_one_fold_by_default(fixture_sample_dataset, fixture_test_strategy):
    """That the splitting strategy only yields one set of splits by default."""
    dataset = fixture_sample_dataset
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset)
    assert len(list(indices)) == 1


def test_can_yield_multiple_folds(fixture_sample_dataset, fixture_test_strategy):
    """That the splitting strategy can generate multiple folds."""
    dataset = fixture_sample_dataset
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, n_splits=2)
    assert len(list(indices)) == 2


def test_default_split_fractions(fixture_sample_dataset, fixture_test_strategy):
    """That the dataset is split by default into 80:10:10 splits."""
    dataset = fixture_sample_dataset
    n_samples = len(dataset)
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset)
    train_indices, validation_indices, test_indices = next(indices)
    assert len(list(train_indices)) == 0.8 * n_samples
    assert len(list(validation_indices)) == 0.1 * n_samples
    assert len(list(test_indices)) == 0.1 * n_samples


def test_custom_split_fractions(fixture_sample_dataset, fixture_test_strategy):
    """That the dataset can be split into arbitrary sizes."""
    dataset = fixture_sample_dataset
    n_samples = len(dataset)
    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=dataset,
        train_fraction=0.6,
        validation_fraction=0.3,
        test_fraction=0.1,
    )
    train_indices, validation_indices, test_indices = next(indices)
    assert len(list(train_indices)) == 0.6 * n_samples
    assert len(list(validation_indices)) == 0.3 * n_samples
    assert len(list(test_indices)) == 0.1 * n_samples


def test_deterministic_split(fixture_sample_dataset, fixture_test_strategy):
    """That split results are deterministic."""
    dataset = fixture_sample_dataset
    strategy = fixture_test_strategy

    indices = strategy.split(dataset=dataset, n_splits=2)
    train_indices1, validation_indices1, test_indices1 = next(indices)
    train_indices2, validation_indices2, test_indices2 = next(indices)

    assert all(i == j for i, j in zip(train_indices1, train_indices2, strict=False))
    assert all(
        i == j for i, j in zip(validation_indices1, validation_indices2, strict=False)
    )
    assert all(i == j for i, j in zip(test_indices1, test_indices2, strict=False))


def test_splits_are_disjoint(fixture_sample_dataset, fixture_test_strategy):
    """That data is spread across splits without overlap."""
    dataset = fixture_sample_dataset
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset)
    train_indices, validation_indices, test_indices = next(indices)
    assert set(train_indices).isdisjoint(set(validation_indices))
    assert set(train_indices).isdisjoint(set(test_indices))
    assert set(validation_indices).isdisjoint(set(test_indices))


def test_splits_are_linear(fixture_test_strategy):
    """That data is split linearly."""
    dataset = list(range(10))
    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=dataset,
        train_fraction=0.6,
        validation_fraction=0.3,
        test_fraction=0.1,
    )
    train_indices, validation_indices, test_indices = next(indices)
    assert list(train_indices) == [0, 1, 2, 3, 4, 5]
    assert list(validation_indices) == [6, 7, 8]
    assert list(test_indices) == [9]


def test_inconsistent_split_fractions_raise(
    fixture_sample_dataset,
    fixture_test_strategy,
):
    dataset = fixture_sample_dataset
    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=dataset,
        train_fraction=0.6,
        validation_fraction=0.1,
        test_fraction=0.1,
    )
    with pytest.raises(AssertionError):
        next(indices)



import os
from typing import Any


def set_test_cache() -> None:
    """Sets a dedicated huggingface cache directory for tests."""
    cache_dir = os.path.expanduser("~/.cache/huggingface/datasets/pytest")
    os.environ["HF_DATASETS_CACHE"] = cache_dir


def mock_load_dataset_to_force_redownload():
    """Mocks the 'datasets.load_dataset()' function called by
    'molflux.datasets.load_dataset()' to force the redownload of assets,
    keeping tests isolated.

    This is because there is no easy way to stop huggingface from downloading
    datasets globally.
    """
    import molflux.datasets

    original_function = molflux.datasets.load.datasets.load_dataset

    def mock_load_dataset(*args: Any, **kwargs: Any):  # type:ignore[no-untyped-def]
        kwargs.update(
            {"download_mode": "force_redownload", "verification_mode": "no_checks"},
        )
        return original_function(*args, **kwargs)

    molflux.datasets.load.datasets.load_dataset = mock_load_dataset


set_test_cache()
mock_load_dataset_to_force_redownload()



"""Expected calibration error"""

import logging
from typing import Any

import evaluate
import numpy as np
from scipy.stats import norm

import datasets
from molflux.metrics.bases import UncertaintyMetric
from molflux.metrics.typing import ArrayLike, MetricResult
from molflux.metrics.uncertainty.utils import _estimate_standard_deviation

logger = logging.getLogger(__name__)


_DESCRIPTION = """
Expected calibration error
The expected calibration error (or ECE) for regression data measures the average difference
between the observed confidence level and the expected confidence level within each bin
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: array of mean predictions.
    references: Ground truth (correct) target values.
    standard_deviations: Standard deviation predicted from a model with uncertainty.
        Either standard deviations or prediction_intervals should be provided.
    prediction_intervals: Prediction intervals of lower and upper bounds.
        Either standard deviations or prediction_intervals should be provided.
    confidence (optional): float indicating what proportion of data lies between
        lower and upper bounds. Defaults to 0.9.
    num_thresholds (optional): int giving the number of thresholds at which to
        assess the calibration error 20. Defaults to 20.

Returns:
    expected_calibration_error: The expected calibration error (ECE) value.

Examples:
    >>> from molflux.metrics import load_metric
    >>> m = load_metric("expected_calibration_error")
    >>> ref = [0, 1, 2, 3, 4]
    >>> pred = [0.5, 0.8, 1.2, 2.4, 4.5]
    >>> lower_bound = [0.1, 0.1, 0.2, 2, 3]
    >>> upper_bound = [1, 1.5, 1.8, 2.8, 5.7]
    >>> prediction_intervals = list(zip(lower_bound, upper_bound))
    >>> m.compute(predictions=pred, references=ref, prediction_intervals=prediction_intervals)
    {'expected_calibration_error': 0.02595}
"""

_CITATION = """\
Naeini, Mahdi Pakdaman, Gregory Cooper, and Milos Hauskrecht.
"Obtaining well calibrated probabilities using bayesian binning."
Twenty-Ninth AAAI Conference on Artificial Intelligence. 2015.
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class ExpectedCalibrationError(UncertaintyMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float64"),
                    "references": datasets.Value("float64"),
                },
            ),
            reference_urls=["https://arxiv.org/pdf/2306.10060.pdf"],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        confidence: float = 0.9,
        num_thresholds: int = 100,
        **kwargs: Any,
    ) -> MetricResult:
        if standard_deviations is None:
            if prediction_intervals is None:
                raise ValueError(
                    "Please provide either standard deviation, or prediction intervals",
                )
            lower_bound, upper_bound = zip(*prediction_intervals, strict=False)
            standard_deviations = _estimate_standard_deviation(
                lower_bound,
                upper_bound,
                confidence,
            )

        assert len(predictions) == len(references)
        pred_cdf = norm.cdf(references, loc=predictions, scale=standard_deviations)
        # expected quantile
        rho_s = [s / num_thresholds for s in range(num_thresholds)]
        # empirical frequency of predicted values falling in the rho_s quantile
        p_s = [np.mean(pred_cdf < rho) for rho in rho_s]
        squared_differences = [
            (rho - p) ** 2 for rho, p in zip(rho_s, p_s, strict=False)
        ]
        score = np.mean(squared_differences)
        return {self.tag: score}



from dataclasses import dataclass
from typing import Any, ClassVar, Literal

import h5py

import datasets
from molflux.datasets.typing import ExamplesGenerator

_BASE_URL = "https://figshare.com/ndownloader/files/18112775"

_HOMEPAGE = "https://www.nature.com/articles/s41597-020-0473-z"

_DESCRIPTION = """
    The ANI-1x and ANI-1ccx ML-based general-purpose potentials for organic molecules
    were developed through active learning; an automated data diversification process.
    The ANI-1x data set contains multiple QM properties from 5 M density
    functional theory calculations, while the ANI-1ccx data set contains 500 k data
    points obtained with an accurate CCSD(T)/CBS extrapolation. Approximately 14 million
    CPU core-hours were expended to generate this data. Multiple QM calculated properties
    for the chemical molfluxs C, H, N, and O are provided: energies, atomic forces, multipole
    moments, atomic charges, etc.

    WARNING: The molecules here are point clouds. The OpenEye mols do not have any bonds.
    """


@dataclass
class ANI1XConfig(datasets.BuilderConfig):
    backend: Literal["openeye", "rdkit"] = "rdkit"


class ANI1X(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = ANI1XConfig
    config: ANI1XConfig

    BUILDER_CONFIGS: ClassVar[list[datasets.BuilderConfig]] = [
        ANI1XConfig(
            name="openeye",
            backend="openeye",
        ),
        ANI1XConfig(
            name="rdkit",
            backend="rdkit",
        ),
    ]

    def _info(self) -> datasets.DatasetInfo:
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "mol_bytes": datasets.Value("binary"),
                    "chemical_formula": datasets.Value("string"),
                    "ccsd(t)_cbs.energy": datasets.Value("float64"),
                    "hf_dz.energy": datasets.Value("float64"),
                    "hf_qz.energy": datasets.Value("float64"),
                    "hf_tz.energy": datasets.Value("float64"),
                    "mp2_dz.corr_energy": datasets.Value("float64"),
                    "mp2_qz.corr_energy": datasets.Value("float64"),
                    "mp2_tz.corr_energy": datasets.Value("float64"),
                    "npno_ccsd(t)_dz.corr_energy": datasets.Value("float64"),
                    "npno_ccsd(t)_tz.corr_energy": datasets.Value("float64"),
                    "tpno_ccsd(t)_dz.corr_energy": datasets.Value("float64"),
                    "wb97x_dz.cm5_charges": datasets.Sequence(
                        datasets.Value("float64"),
                    ),
                    "wb97x_dz.dipole": datasets.Sequence(
                        datasets.Value("float64"),
                        length=-1,
                    ),
                    "wb97x_dz.energy": datasets.Value("float64"),
                    "wb97x_dz.forces": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Value(dtype="float64", id=None),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "wb97x_dz.hirshfeld_charges": datasets.Sequence(
                        datasets.Value("float64"),
                    ),
                    "wb97x_dz.quadrupole": datasets.Sequence(
                        datasets.Value("float64"),
                        length=-1,
                    ),
                    "wb97x_tz.dipole": datasets.Sequence(
                        datasets.Value("float64"),
                        length=-1,
                    ),
                    "wb97x_tz.energy": datasets.Value("float64"),
                    "wb97x_tz.forces": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Value(dtype="float64", id=None),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "wb97x_tz.mbis_charges": datasets.Sequence(
                        datasets.Value("float64"),
                    ),
                    "wb97x_tz.mbis_dipoles": datasets.Sequence(
                        datasets.Value("float64"),
                    ),
                    "wb97x_tz.mbis_octupoles": datasets.Sequence(
                        datasets.Value("float64"),
                    ),
                    "wb97x_tz.mbis_quadrupoles": datasets.Sequence(
                        datasets.Value("float64"),
                    ),
                    "wb97x_tz.mbis_volumes": datasets.Sequence(
                        datasets.Value("float64"),
                    ),
                },
            ),
            homepage=_HOMEPAGE,
        )

    def _split_generators(
        self,
        dl_manager: datasets.DownloadManager,
    ) -> list[datasets.SplitGenerator]:
        archive_path = dl_manager.download_and_extract(_BASE_URL)

        return [
            datasets.SplitGenerator(
                name="train",
                gen_kwargs={
                    "h5_file_path": archive_path,
                },
            ),
        ]

    def _generate_examples(self, **kwargs: Any) -> ExamplesGenerator:
        if self.config.backend == "openeye":
            return self._generate_examples_openeye(**kwargs)
        elif self.config.backend == "rdkit":
            return self._generate_examples_rdkit(**kwargs)
        else:
            raise KeyError("Unknown backend")

    def _generate_examples_openeye(
        self,
        h5_file_path: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from openeye.oechem import OEMol, OEWriteMolToBytes
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("openeye-toolkits", e) from e

        # open h5 file
        h5_file = h5py.File(h5_file_path, "r+")

        # get columns to loop over
        columns = list(self._info().features.keys())
        columns.remove("mol_bytes")
        columns.append("coordinates")

        # reset index
        index = 0

        # loop over chemical formulae (each have multiple conformers)
        for chemical_formula in h5_file:
            # the dict of data for this chemical formula
            chemical_formula_dict = {
                key: value[:].tolist()
                for key, value in dict(h5_file[chemical_formula]).items()
            }

            # atomic numbers list for the chemical formula
            atomic_nums = chemical_formula_dict["atomic_numbers"]
            del chemical_formula_dict["atomic_numbers"]

            # make all the molecules into mol bytes (only atomic numbers and coords available)
            mol_list = []
            for coords in chemical_formula_dict["coordinates"]:
                mol = OEMol()
                for atom_num in atomic_nums:
                    mol.NewAtom(atom_num)
                flat_coords = [xs for x in coords for xs in x]
                mol.SetCoords(flat_coords)

                mol_bytes = OEWriteMolToBytes(".oeb", mol)
                mol_list.append(mol_bytes)

            # delete coordinates from the dict and add mols
            del chemical_formula_dict["coordinates"]
            chemical_formula_dict["mol_bytes"] = mol_list

            # find list of example dicts (swap dict of lists to list of dicts)
            chemical_formula_examples_list = [
                dict(zip(chemical_formula_dict, t, strict=False))
                for t in zip(*chemical_formula_dict.values(), strict=False)
            ]

            # yield the examples for the conformers
            for example_dict in chemical_formula_examples_list:
                index += 1

                # add chemical formula
                example_dict["chemical_formula"] = chemical_formula
                yield index, example_dict

    def _generate_examples_rdkit(
        self,
        h5_file_path: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from rdkit.Chem import Atom, Conformer, RWMol
            from rdkit.Geometry import Point3D
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("rdkit", e) from e

        # open h5 file
        h5_file = h5py.File(h5_file_path, "r+")

        # get columns to loop over
        columns = list(self._info().features.keys())
        columns.remove("mol_bytes")
        columns.append("coordinates")

        # reset index
        index = 0

        # loop over chemical formulae (each have multiple conformers)
        for chemical_formula in h5_file:
            # the dict of data for this chemical formula
            chemical_formula_dict = {
                key: value[:].tolist()
                for key, value in dict(h5_file[chemical_formula]).items()
            }

            # atomic numbers list for the chemical formula
            atomic_nums = chemical_formula_dict["atomic_numbers"]
            del chemical_formula_dict["atomic_numbers"]

            # make all the molecules into mol bytes (only atomic numbers and coords available)
            mol_list = []
            for coords in chemical_formula_dict["coordinates"]:
                mol = RWMol()
                for _idx, atomic_num in enumerate(atomic_nums):
                    atom = Atom(atomic_num)
                    mol.AddAtom(atom)

                conf = Conformer()
                for i in range(mol.GetNumAtoms()):
                    conf.SetAtomPosition(i, Point3D(*coords[i]))

                mol.AddConformer(conf)

                mol_list.append(mol.ToBinary())

            # delete coordinates from the dict and add mols
            del chemical_formula_dict["coordinates"]
            chemical_formula_dict["mol_bytes"] = mol_list

            # find list of example dicts (swap dict of lists to list of dicts)
            chemical_formula_examples_list = [
                dict(zip(chemical_formula_dict, t, strict=False))
                for t in zip(*chemical_formula_dict.values(), strict=False)
            ]

            # yield the examples for the conformers
            for example_dict in chemical_formula_examples_list:
                index += 1

                # add chemical formula
                example_dict["chemical_formula"] = chemical_formula
                yield index, example_dict



import logging
from typing import Any

try:
    from openeye import oechem, oeshape
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import to_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

logger = logging.getLogger(__name__)

_DESCRIPTION = """
A representation based on coefficients of a hermite polynomial expansion, as used in methods such as ROCS
Details on hermite expansions and shape theory for molecules can be found here:
https://docs.eyesopen.com/toolkits/python/shapetk/shape_theory.html
"""


class Hermite(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        n_poly_max: int = 8,
        use_optimal_lambdas: bool = True,
        orient_by_moments_of_inertia: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[list[float]]]:
        r"""Featurises the input molecules via a hermite polynomial expansion.

        Args:
            n_poly_max: Controls the level of Hermite expansion. In particular
                the lowest value 0 corresponds to a very inaccurate expansion,
                approximating the entire molecule by a single Gaussian, while
                the largest currently allowed value is equal to 30, which
                corresponds to a list of 5456 Hermite coefficients. In the limit
                `n_poly_max` tending to infinity, we obtain exact equivalence
                of the Hermite expansion to the Gaussian representation of the
                molecule. Defaults to `8`.
            use_optimal_lambdas: If `True`, optimization of lambda parameters
                will be performed via maximizing the self-overlap of each input
                molecule at a given resolution of the Hermite expansion. Please
                note, that even in the case when set to `False`, if the value
                of `n_poly_max` is high enough the Hermite expansion will
                converge to the Gaussian shape. However, in practice at a given
                low resolution of Hermite expansion it is essential to use the
                optimal lambda parameters for faster convergence. Defaults to
                `True`.
            orient_by_moments_of_inertia: If `True`, each input sample is moved
                to the center of its inertial frame before calculating its
                Hermite coefficients. Defaults to `False`.

        Returns:
            Coefficients of a hermite expansion representing the shape of a molecule

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('hermite')
            >>> samples = ['C']
            >>> representation.featurise(samples, n_poly_max=1)
            {'hermite': [[2.70..., 0.0, 0.0, 0.0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        hermite_mols = []
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                out = _get_hermite_coeffs(
                    mol,
                    n_poly_max=n_poly_max,
                    use_optimal_lambdas=use_optimal_lambdas,
                    orient_by_moments_of_inertia=orient_by_moments_of_inertia,
                )
                hermite_mols.append(list(out))

        return {self.tag: hermite_mols}


def _get_hermite_coeffs(
    mol: oechem.OEMolBase,
    n_poly_max: int,
    *,
    use_optimal_lambdas: bool = True,
    orient_by_moments_of_inertia: bool = False,
) -> oechem.OEDoubleVector:
    """Calculates Hermite coefficients."""

    opts = oeshape.OEHermiteOptions()
    opts.SetNPolyMax(n_poly_max)
    opts.SetUseOptimalLambdas(use_optimal_lambdas)
    hermite = oeshape.OEHermite(opts)

    if orient_by_moments_of_inertia:
        trans = oechem.OETrans()
        oeshape.OEOrientByMomentsOfInertia(mol, trans)

    if not hermite.Setup(mol):
        logger.error("failed to setup molecule {} for OEHermite.", mol.GetTitle())

    # Length of the output vector can be determined as number of terms that meet
    # the condition l+m+n < n_poly_max. See https://docs.eyesopen.com/toolkits/python/shapetk/OEShapeClasses/OEHermite.html#OEShape::OEHermite::GetCoefficients
    basis_size = (n_poly_max + 1) * (n_poly_max + 2) * (n_poly_max + 3) / 6
    basis_size = int(basis_size)

    coeffs = oechem.OEDoubleVector(basis_size)
    hermite.GetCoefficients(coeffs)

    return coeffs



from collections.abc import Iterable
from typing import Any, Literal

import numpy as np
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)

try:
    from sklearn.discriminant_analysis import (
        LinearDiscriminantAnalysis as SKLinearDiscriminantAnalysis,
    )
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
Linear Discriminant Analysis.

A classifier with a linear decision boundary, generated by fitting class
conditional densities to the data and using Bayes' rule.

The model fits a Gaussian density to each class, assuming that all classes
share the same covariance matrix.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
solver : {'svd', 'lsqr', 'eigen'}, default='svd'
    Solver to use, possible values:
      - 'svd': Singular value decomposition (default).
        Does not compute the covariance matrix, therefore this solver is
        recommended for data with a large number of features.
      - 'lsqr': Least squares solution.
        Can be combined with shrinkage or custom covariance estimator.
      - 'eigen': Eigenvalue decomposition.
        Can be combined with shrinkage or custom covariance estimator.
shrinkage : 'auto' or float, default=None
    Shrinkage parameter, possible values:
      - None: no shrinkage (default).
      - 'auto': automatic shrinkage using the Ledoit-Wolf lemma.
      - float between 0 and 1: fixed shrinkage parameter.
    This should be left to None if `covariance_estimator` is used.
    Note that shrinkage works only with 'lsqr' and 'eigen' solvers.
priors : array-like of shape (n_classes,), default=None
    The class prior probabilities. By default, the class proportions are
    inferred from the training data.
n_components : int, default=None
    Number of components (<= min(n_classes - 1, n_features)) for
    dimensionality reduction. If None, will be set to
    min(n_classes - 1, n_features). This parameter only affects the
    `transform` method.
store_covariance : bool, default=False
    If True, explicitly compute the weighted within-class covariance
    matrix when solver is 'svd'. The matrix is always computed
    and stored for the other solvers.
tol : float, default=1.0e-4
    Absolute threshold for a singular value of X to be considered
    significant, used to estimate the rank of X. Dimensions whose
    singular values are non-significant are discarded. Only used if
    solver is 'svd'.
covariance_estimator : covariance estimator, default=None
    If not None, `covariance_estimator` is used to estimate
    the covariance matrices instead of relying on the empirical
    covariance estimator (with potential shrinkage).
    The object should have a fit method and a ``covariance_`` attribute
    like the estimators in :mod:`sklearn.covariance`.
    if None the shrinkage parameter drives the estimate.
    This should be left to None if `shrinkage` is used.
    Note that `covariance_estimator` works only with 'lsqr' and 'eigen'
    solvers.
"""

SolverT = Literal["svd", "lsqr", "eigen"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class LinearDiscriminantAnalysisClassifierConfig(ModelConfig):
    solver: SolverT = "svd"
    shrinkage: None | int | Literal["auto"] = None
    priors: Iterable[float] | None = None
    n_components: int | None = None
    store_covariance: bool = False
    tol: float = 1e-4
    covariance_estimator: Any | None = None


class LinearDiscriminantAnalysisClassifier(
    SKLearnClassificationMixin,
    SKLearnModelBase[LinearDiscriminantAnalysisClassifierConfig],
):
    @property
    def _config_builder(self) -> type[LinearDiscriminantAnalysisClassifierConfig]:
        return LinearDiscriminantAnalysisClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKLinearDiscriminantAnalysis:
        config = self.model_config
        return SKLinearDiscriminantAnalysis(
            solver=config.solver,
            shrinkage=config.shrinkage,
            priors=None if config.priors is None else np.array(config.priors),
            n_components=config.n_components,
            store_covariance=config.store_covariance,
            tol=config.tol,
            covariance_estimator=config.covariance_estimator,
        )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

import pytest
from cloudpathlib import implementation_registry
from cloudpathlib.local import LocalS3Client, local_s3_implementation


@pytest.fixture
def monkeypatched_aws_credentials(monkeypatch):
    """Mocked AWS Credentials to make sure we are not accessing real resources."""
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "testing")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "testing")
    monkeypatch.setenv("AWS_SECURITY_TOKEN", "testing")
    monkeypatch.setenv("AWS_SESSION_TOKEN", "testing")
    monkeypatch.setenv("AWS_DEFAULT_REGION", "eu-west-1")


@pytest.fixture
def fixture_empty_bucket(monkeypatch, monkeypatched_aws_credentials):
    """Fixture that patches CloudPath dispatch and also sets up test assets in LocalS3Client's
    local storage directory.

    References:
        https://cloudpathlib.drivendata.org/stable/testing_mocked_cloudpathlib/
    """

    monkeypatch.setitem(implementation_registry, "s3", local_s3_implementation)

    mock_bucket: Path = LocalS3Client.get_default_storage_dir() / "mock-bucket"
    mock_bucket.mkdir(exist_ok=True, parents=True)

    yield mock_bucket

    LocalS3Client.reset_default_storage_dir()  # clean up temp directory and replace with new one



import contextlib
from collections.abc import Callable, Iterator
from types import FunctionType
from typing import Any

from molflux.features.errors import FeaturisationError, InvalidNumberOfPositionalArgs


def copyfunc(func: Callable) -> FunctionType:
    """Makes a complete copy of a given function.

    References:
        https://github.com/huggingface/datasets/blob/95193ae61e92aa537d0c65d37a1fd9d2393aae89/src/datasets/utils/py_utils.py#L681
    """
    result = FunctionType(
        func.__code__,
        func.__globals__,
        func.__name__,
        func.__defaults__,
        func.__closure__,
    )
    result.__kwdefaults__ = func.__kwdefaults__
    return result


@contextlib.contextmanager
def featurisation_error_harness(sample: Any) -> Iterator:
    """Creates a context within which a FeaturisationError is raised on any Exception.

    Args:
        sample: The sample to wrap.

    Raises:
        molflux.features.errors.FeaturisationError: If any Exception is raised within the context.

    Examples:

        >>> iterable = [1, 2, 3]
        >>> for x in iterable:  # doctest: +SKIP
        >>>     with featurisation_error_harness(x):  # doctest: +SKIP
        >>>         # do something that might raise
    """
    try:
        yield
    except Exception as e:
        raise FeaturisationError(sample=sample) from e


def assert_n_positional_args(*args: Any, expected_size: int) -> None:
    """Checks that the number of positional arguments given matches the expected size.

    Args:
        *args: The positional arguments whose number is checked
        expected_size: The expected number of positional arguments

    Raises:
            molflux.features.errors.InvalidNumberOfPositionalArgs: If the number of args is different from expected_size.

    """
    actual_size = len(args)
    if actual_size != expected_size:
        raise InvalidNumberOfPositionalArgs(expected_size, actual_size)



import pytest
import torch
from pydantic.v1 import ValidationError
from torch._dynamo.testing import CompileCounter

from molflux.modelzoo import save_to_store
from molflux.modelzoo.models.lightning.config import (
    CompileConfig,
    TransferLearningStage,
)
from molflux.modelzoo.models.lightning.model import ConfigOverride


def test_transfer_learning_stage_misspelt_kwarg():
    with pytest.raises(ValidationError):
        TransferLearningStage(trainer={"accele": "cpu"})


@pytest.mark.parametrize(
    "compile",
    [True, False],
)
def test_transfer_learning_model(
    tmp_path,
    fixture_model,
    fixture_pre_trained_model,
    train_dataset,
    validation_dataset,
    compile,
    compile_error_context,
):
    """Test that transfer learning for one stage trains the right params and has correct weights"""

    model = fixture_pre_trained_model
    save_to_store(str(tmp_path), model)

    new_model = fixture_model
    with compile_error_context:
        new_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 1,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
                "validation": {"batch_size": 2},
            },
            transfer_learning_config={
                "pre_trained_model_path": str(tmp_path),
                "stages": [
                    {
                        "freeze_modules": [
                            "module.0",
                        ],
                    },
                ],
            },
            compile_config=compile,
        )

        module = new_model.module if not compile else new_model.module._orig_mod

        for n, p in module.named_parameters():
            # check that matched and frozen is equal to
            if n == "module.0.weight":
                assert torch.equal(p, torch.ones_like(p) * 0)
            elif n == "module.0.bias":
                assert torch.equal(p, torch.ones_like(p) * 1)
            elif n == "module.2.weight":
                assert (not torch.equal(p, torch.ones_like(p) * 2)) and (
                    torch.allclose(p, torch.ones_like(p) * 2, atol=1e-3)
                )
            elif n == "module.2.bias":
                assert (not torch.equal(p, torch.ones_like(p) * 3)) and (
                    torch.allclose(p, torch.ones_like(p) * 3, atol=1e-3)
                )
            else:
                raise KeyError("Param not found in module")


def test_transfer_learning_model_overrides(fixture_model):
    """Test that transfer learning stage overrides work properly"""

    model = fixture_model

    config_override = ConfigOverride(
        model,
        transfer_learning={
            "pre_trained_model_path": "dummy_path",
            "stages": [
                {
                    "trainer": {
                        "max_epochs": 42,
                    },
                    "datamodule": {
                        "train": {"batch_size": 42},
                    },
                    "optimizer": {
                        "name": "AdamW",
                        "config": {
                            "lr": 42.0,
                        },
                    },
                    "scheduler": {
                        "config": {
                            "T_max": 10,
                        },
                    },
                },
            ],
        },
    )

    with config_override:
        with ConfigOverride(
            model,
            trainer=model.model_config.transfer_learning.stages[0].trainer,
            datamodule=model.model_config.transfer_learning.stages[0].datamodule,
            optimizer=model.model_config.transfer_learning.stages[0].optimizer,
            scheduler=model.model_config.transfer_learning.stages[0].scheduler,
        ):
            assert model.model_config.trainer.max_epochs == 42
            assert model.model_config.datamodule.train.batch_size == 42
            assert model.model_config.optimizer.name == "AdamW"
            assert model.model_config.optimizer.config["lr"] == 42.0
            assert model.model_config.scheduler.name == "CosineAnnealingLR"
            assert model.model_config.scheduler.config["T_max"] == 10


@pytest.mark.parametrize(
    "compile",
    [True, False],
)
def test_transfer_learning_model_partial_match(
    tmp_path,
    fixture_model,
    fixture_pre_trained_model,
    train_dataset,
    validation_dataset,
    compile,
    compile_error_context,
):
    """Test that transfer learning for one stage trains the right params and has correct weights"""

    model = fixture_pre_trained_model
    save_to_store(str(tmp_path), model)

    new_model = fixture_model
    with compile_error_context:
        new_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 1,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
                "validation": {"batch_size": 2},
            },
            transfer_learning_config={
                "pre_trained_model_path": str(tmp_path),
                "modules_to_match": {"module.0": "module.0"},
                "stages": [
                    {
                        "freeze_modules": [
                            "module.2",
                        ],
                    },
                ],
            },
            compile_config=compile,
        )

        module = new_model.module if not compile else new_model.module._orig_mod

        for n, p in module.named_parameters():
            # check that matched but not frozen is close but not equal to
            if n == "module.0.weight":
                assert (not torch.equal(p, torch.ones_like(p) * 0)) and (
                    torch.allclose(p, torch.ones_like(p) * 0, atol=1e-3)
                )
            elif n == "module.0.bias":
                assert (not torch.equal(p, torch.ones_like(p) * 1)) and (
                    torch.allclose(p, torch.ones_like(p) * 1, atol=1e-3)
                )
            elif n == "module.2.weight":
                assert (not torch.equal(p, torch.ones_like(p) * 2)) and (
                    not torch.allclose(p, torch.ones_like(p) * 2, atol=1e-3)
                )
            elif n == "module.2.bias":
                assert (not torch.equal(p, torch.ones_like(p) * 3)) and (
                    not torch.allclose(p, torch.ones_like(p) * 3, atol=1e-3)
                )
            else:
                raise KeyError("Param not found in module")


@pytest.mark.parametrize(
    "compile",
    [True, False],
)
def test_transfer_learning_model_surjective_match(
    tmp_path,
    fixture_multi_layer_model,
    fixture_pre_trained_multi_layer_model,
    train_dataset,
    validation_dataset,
    compile,
    compile_error_context,
):
    """Test that transfer learning for one stage with surjective matching trains the right params and has correct weights"""

    model = fixture_pre_trained_multi_layer_model
    save_to_store(str(tmp_path), model)

    new_model = fixture_multi_layer_model
    with compile_error_context:
        new_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 1,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
                "validation": {"batch_size": 2},
            },
            transfer_learning_config={
                "pre_trained_model_path": str(tmp_path),
                "modules_to_match": {
                    "module.2": "module.2",
                    "module.4": "module.2",
                },
                "stages": [
                    {
                        "freeze_modules": [
                            "module.2",
                        ],
                    },
                ],
            },
            compile_config=compile,
        )

        module = new_model.module if not compile else new_model.module._orig_mod

        for n, p in module.named_parameters():
            # check that matched but not frozen is close but not equal to
            if n == "module.2.weight":
                assert torch.equal(p, torch.ones_like(p) * 0)
            elif n == "module.2.bias":
                assert torch.equal(p, torch.ones_like(p) * 1)
            elif n == "module.4.weight":
                assert (not torch.equal(p, torch.ones_like(p) * 0)) and (
                    torch.allclose(p, torch.ones_like(p) * 0, atol=1e-3)
                )
            elif n == "module.4.bias":
                assert (not torch.equal(p, torch.ones_like(p) * 1)) and (
                    torch.allclose(p, torch.ones_like(p) * 1, atol=1e-3)
                )


@pytest.mark.parametrize(
    "compile",
    [True, False],
)
def test_transfer_learning_model_multistage(
    tmp_path,
    fixture_model,
    fixture_pre_trained_model,
    train_dataset,
    validation_dataset,
    compile,
    compile_error_context,
):
    """Test that transfer learning for multistages trains the right params and has correct weights"""

    model = fixture_pre_trained_model
    save_to_store(str(tmp_path), model)

    new_model = fixture_model
    with compile_error_context:
        new_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 1,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
                "validation": {"batch_size": 2},
            },
            transfer_learning_config={
                "pre_trained_model_path": str(tmp_path),
                "stages": [
                    {
                        "freeze_modules": [
                            "module.0",
                        ],
                    },
                    {
                        "freeze_modules": [
                            "module.2",
                        ],
                    },
                ],
            },
            compile_config=compile,
        )

        module = new_model.module if not compile else new_model.module._orig_mod

        for n, p in module.named_parameters():
            if n == "module.0.weight":
                assert (not torch.equal(p, torch.ones_like(p) * 0)) and (
                    torch.allclose(p, torch.ones_like(p) * 0, atol=1e-3)
                )
            elif n == "module.0.bias":
                assert (not torch.equal(p, torch.ones_like(p) * 1)) and (
                    torch.allclose(p, torch.ones_like(p) * 1, atol=1e-3)
                )
            elif n == "module.2.weight":
                assert (not torch.equal(p, torch.ones_like(p) * 2)) and (
                    torch.allclose(p, torch.ones_like(p) * 2, atol=1e-3)
                )
            elif n == "module.2.bias":
                assert (not torch.equal(p, torch.ones_like(p) * 3)) and (
                    torch.allclose(p, torch.ones_like(p) * 3, atol=1e-3)
                )
            else:
                raise KeyError("Param not found in module")


@pytest.mark.parametrize(
    "compile",
    [True, False],
)
def test_transfer_learning_model_bad_module_name(
    tmp_path,
    fixture_model,
    fixture_pre_trained_model,
    train_dataset,
    validation_dataset,
    compile,
    compile_error_context,
):
    """Test that transfer learning errors on bad module name"""
    model = fixture_pre_trained_model
    save_to_store(str(tmp_path), model)

    new_model = fixture_model

    with pytest.raises(KeyError), compile_error_context:
        new_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 1,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
                "validation": {"batch_size": 2},
            },
            transfer_learning_config={
                "pre_trained_model_path": str(tmp_path),
                "modules_to_match": {"bad_name": "module.0"},
                "stages": [
                    {
                        "freeze_modules": [
                            "module.0",
                        ],
                    },
                ],
            },
            compile_config=compile,
        )

    with pytest.raises(KeyError), compile_error_context:
        new_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 1,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
            },
            transfer_learning_config={
                "pre_trained_model_path": str(tmp_path),
                "modules_to_match": {"module.0": "bad_name"},
                "stages": [
                    {
                        "freeze_modules": [
                            "module.0",
                        ],
                    },
                ],
            },
            compile_config=compile,
        )

    with pytest.raises(KeyError), compile_error_context:
        new_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 1,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
            },
            transfer_learning_config={
                "pre_trained_model_path": str(tmp_path),
                "stages": [
                    {
                        "freeze_modules": [
                            "bad_name",
                        ],
                    },
                ],
            },
            compile_config=compile,
        )


@pytest.mark.parametrize("compile", [True, False])
def test_does_compile(
    tmp_path,
    fixture_model,
    fixture_pre_trained_model,
    train_dataset,
    validation_dataset,
    compile_error_context,
    compile,
):
    """Test that the model does compile during training when set to"""
    # Python 3.11 does not support compilation as of PyTorch 2.0.1, but 2.1 will

    model = fixture_pre_trained_model
    save_to_store(str(tmp_path), model)

    new_model = fixture_model

    if compile:
        cnt = CompileCounter()
        compile_config = CompileConfig("default")
        new_model.model_config.compile = compile_config  # pyright: ignore
        new_model.model_config.compile.backend = cnt  # pyright: ignore

        assert cnt.frame_count == 0

    assert not new_model.is_compiled

    with compile_error_context:
        new_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 1,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
                "validation": {"batch_size": 2},
            },
            transfer_learning_config={
                "pre_trained_model_path": str(tmp_path),
                "stages": [
                    {
                        "freeze_modules": [
                            "module.0",
                        ],
                    },
                ],
            },
        )

        if compile:
            assert cnt.frame_count > 0  # pyright: ignore
            assert new_model.is_compiled



from __future__ import annotations

import ctypes
import difflib
import logging
from collections.abc import Callable, Iterable, Iterator
from typing import Any

import numpy as np

try:
    from openeye import oechem, oegraphsim
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

logger = logging.getLogger(__name__)


def oemol_from_smiles(smiles: str) -> oechem.OEMolBase:
    """Returns a OEMol object representing the input SMILES string."""
    oemol = oechem.OEMol()
    if oechem.OESmilesToMol(oemol, smiles):
        return oemol


def oemol_from_bytes(bytes_molecule: bytes) -> oechem.OEMolBase:
    """Returns a OEMol object representing the input bytes string."""
    oemol = oechem.OEMol()
    if oechem.OEReadMolFromBytes(oemol, ".oeb", bytes_molecule):
        return oemol


def oemol_from_hex(hex_molecule: str) -> oechem.OEMolBase:
    """Returns a OEMol object representing the input hex string."""
    bytes_molecule = bytes.fromhex(hex_molecule)
    return oemol_from_bytes(bytes_molecule)


def smiles_from_oemol(oemol: oechem.OEMolBase, flavor: int | None = None) -> str:
    """
    Returns a SMILES string representing the input OEMol object.
    """
    if flavor is None:
        flavor = oechem.OEGetDefaultOFlavor(oechem.OEFormat_SMI)
    smiles: str = oechem.OEWriteMolToString(oechem.OEFormat_SMI, flavor, False, oemol)
    return smiles.rstrip("\n")


def smiles_from_bytes(bytes_molecule: bytes) -> str:
    """
    Returns a SMILES string representing the input bytes string.
    """
    oemol = oemol_from_bytes(bytes_molecule)
    return smiles_from_oemol(oemol)


def to_oemol(molecule: Any) -> oechem.OEMolBase:
    """
    Converts a single input to an OEMol object. This is done by going over the following cases in order:

    - input is an OEMolBase child, in which case it is returned unchanged
    - input is a string
        - if it can be parsed as a SMILES, it is converted to an OEMol from the SMILES
        - else, if it is a hexadecimal string, it is converted to bytes and then to an OEMol
        - else, raise a TypeError
    - input is a bytes
        - it is converted to an OEMol
    - else, raise an error

    Args:
        molecule: Any

    Returns:
        OEMolBase
    """

    if isinstance(molecule, oechem.OEMolBase):
        return molecule

    if isinstance(molecule, str):
        try:
            oemol = oemol_from_smiles(molecule)
            if oemol is not None:
                return oemol
        except:  # noqa: E722 S110
            pass

        try:
            return oemol_from_hex(molecule)
        except ValueError:
            pass

        raise TypeError(
            f"Unparsable string input sample: {molecule}. Tried to parse as SMILES and HEX.",
        )

    elif isinstance(molecule, bytes):
        return oemol_from_bytes(molecule)

    else:
        raise TypeError(f"Unsupported input sample type: {type(molecule)!r}")


def to_smiles(molecule: Any) -> str:
    """
    Converts a single sample to a SMILES string.
    """

    if isinstance(molecule, str):
        if oechem.OEParseSmiles(
            oechem.OEMol(),
            molecule,
            oechem.OEParseSmilesOptions(quiet=True),
        ):
            return molecule

        try:
            return smiles_from_oemol(oemol_from_hex(molecule))
        except ValueError:
            pass

        raise TypeError(
            f"Unparsable string input sample: {molecule}. Tried to parse as SMILES and HEX.",
        )

    elif isinstance(molecule, oechem.OEMolBase):
        return smiles_from_oemol(molecule)

    elif isinstance(molecule, bytes):
        return smiles_from_bytes(molecule)

    else:
        raise TypeError(f"Unsupported input sample type: {type(molecule)!r}")


def to_hex(molecule: Any) -> str:
    """
    Converts a single sample to bytes and then to a hex string.
    """
    mol = to_oemol(molecule)
    mol_bytes = oechem.OEWriteMolToBytes(".oeb", mol)
    mol_hex = str(mol_bytes.hex())
    return mol_hex


def write_mols(
    mols: Iterable[oechem.OEMolBase],
    output_file: str,
) -> None:
    # dest must be a valid output format
    if not oechem.OEIsWriteable(output_file):
        raise ValueError(
            f"`{output_file}` is not a supported chemical structure format.",
        )

    num_mols = 0
    with oechem.oemolostream(output_file) as oss:
        for mol in mols:
            # written as a copy since OEWriteMolecule can change the object
            oechem.OEWriteMolecule(oss, mol.CreateCopy())
            num_mols += 1


def to_digit(data: Any, start: int, stop: int, num: int) -> np.ndarray:
    bins = np.linspace(start=start, stop=stop, num=num)
    binned = np.digitize([data], bins, right=False)
    digit = np.zeros(bins.size + 1, dtype=int)
    for i in binned:
        digit[i] = 1
    return digit


def iterate_mols_or_confs(
    mols: Iterable[oechem.OEMolBase],
    expand_conformers: bool = False,
) -> Iterator[oechem.OEMolBase]:
    """Iterates over molecules and yields either the molecule if it's an OEGraphMol,
    or the conformers if it's an OEMol.

    Args:
        mols: Iterable of molecule objects.
        expand_conformers: Determines whether to expand out conformations of an
            OEMol object (`True`) or to simply yield the active conformation
            (`False`, default).

    Yields:
        Each molecule or conformation in `mols`.
    """
    for mol in mols:
        if isinstance(mol, oechem.OEMCMolBase):
            yield from mol.GetConfs() if expand_conformers else (mol.GetActive(),)
        else:
            yield mol


def get_sd_data(
    mol: oechem.OEMolBase,
    tag: str,
    *,
    dtype: Callable[[Any], Any] = str,
) -> Any:
    """Function to obtain SD data from a molecule object by `tag`.

    Args:
        mol: The molecule to obtain the SD data for.
        tag: The SD data tag to retrieve.
        dtype: The type to cast the value associated with `tag` to. Default is
            `str`.

    Returns:
        The value from the SD tag of `mol` given by `tag`.

    Notes:
        Note, that if `mol` is a conformer, the `tag` is checked on the parent
        MC molecule object if the tag is not present on the conformer. If the
        `mol` object is an `OEMol` object and the tag is not present it will
        check the active conformer for the presence of the tag (via `.GetActive()`).
    """
    if not oechem.OEHasSDData(mol, tag):
        if isinstance(mol, oechem.OEConfBase) and oechem.OEHasSDData(
            mol.GetMCMol(),
            tag,
        ):
            mol = mol.GetMCMol()
        elif isinstance(mol, oechem.OEMCMolBase) and oechem.OEHasSDData(
            mol.GetActive(),
            tag,
        ):
            mol = mol.GetActive()
        else:
            sd_tags = difflib.get_close_matches(
                tag,
                {
                    sd_data_pair.GetTag()
                    for sd_data_pair in oechem.OEGetSDDataPairs(mol)
                },
                # 0.1 cutoff used to provide users with more possible options.
                # cutoff is based on textual similarity, lower is more lenient.
                cutoff=0.1,
            )
            raise ValueError(
                f"Unable to find SD tag {tag!s} on molecule {mol.GetTitle()}. Closest "
                f"SD tags found: {(*sd_tags,)}",
            )
    value = oechem.OEGetSDData(mol, tag)
    try:
        value = dtype(value)
    except Exception as err:
        raise TypeError(
            f"Error occurred when attempting to cast {tag!s}={value} to {dtype=}. Error: {err}",
        ) from err

    return value


def fingerprint_to_bit_vector(fingerprint: oegraphsim.OEFingerPrint) -> list[int]:
    """Casts a chemical fingerprint to a bit vector."""

    # calculate the number of bytes / fingerprint can have any size
    fingerprint_size = fingerprint.GetSize()
    nbytes: int = (fingerprint_size // 8) + (1 if fingerprint_size % 8 else 0)

    # get fp bytes from c pointer
    array_type = ctypes.POINTER(ctypes.c_ubyte * nbytes)
    pointer = ctypes.cast(int(fingerprint.GetData()), array_type)
    bits = np.unpackbits(np.ctypeslib.as_array(pointer, shape=(1,)))

    # get correct fp length back
    return bits.reshape(nbytes, 8)[:, ::-1].ravel()[:fingerprint_size].tolist()  # type: ignore[no-any-return]



import molflux.splits
from molflux.datasets.interfaces import SplittingStrategy


def test_integrates_with_molflux_splits():
    """That the SplittingStrategy interface used by datasets is compatible
    with the SplittingStrategy protocol defined by molflux.features.

    That is, we check that datasets works with strategies from molflux.features.
    """
    assert isinstance(molflux.splits.SplittingStrategy, SplittingStrategy)



import pytest

from molflux.metrics.parsers import Spec, dict_parser, yaml_parser


def test_dict_parser_raises_syntax_error_on_invalid_dict():
    """That a SyntaxError is raised on validation errors."""
    dictionary = {"wrong": "schema"}
    with pytest.raises(SyntaxError):
        dict_parser(dictionary)


def test_dict_parser_parses_to_spec():
    """That dictionaries get parsed into Spec objects."""
    dictionary = {
        "name": "metric name",
        "config": {},
        "presets": {},
    }
    spec = dict_parser(dictionary)
    assert isinstance(spec, Spec)


def test_yaml_parser_parses_minimal_config(fixture_path_to_assets):
    path = fixture_path_to_assets / "minimal_config.yml"
    specs = yaml_parser(path)
    assert specs


def test_yaml_parser_parses_full_config(fixture_path_to_assets):
    path = fixture_path_to_assets / "config.yml"
    specs = yaml_parser(path)
    assert specs



"""Matthews correlation coefficient (MCC)."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import matthews_corrcoef

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
The Matthews correlation coefficient is used in machine learning as a measure of
the quality of binary and multiclass classifications. It takes into account true
and false positives and negatives and is generally regarded as a balanced
measure which can be used even if the classes are of very different sizes.
The MCC is in essence a correlation coefficient value between -1 and +1. A
coefficient of +1 represents a perfect prediction, 0 an average random
prediction and -1 an inverse prediction. The statistic is also known as the phi
coefficient.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated targets as returned by a classifier.
    references: Ground truth (correct) target values.
    sample_weight (optional): Weighting of each sample.

Returns:
    matthews_corrcoef: The Matthews correlation coefficient (+1 represents a
    perfect prediction, 0 an average random prediction and -1 and inverse
    prediction).

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("matthews_corrcoef")
    >>> references = [+1, +1, +1, -1]
    >>> predictions = [+1, -1, +1, +1]
    >>> metric.compute(references=references, predictions=predictions)
    {'matthews_corrcoef': -0.33...}
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class MatthewsCorrcoef(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("int32"),
                    "references": datasets.Value("int32"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.matthews_corrcoef.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        score = matthews_corrcoef(
            y_true=references,
            y_pred=predictions,
            sample_weight=sample_weight,
        )
        return {self.tag: float(score)}



import numpy as np
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.regression.mean_squared_error import MeanSquaredError


@pytest.fixture()
def fixture_metric():
    return load_metric("mean_squared_error")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "mean_squared_error" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, MeanSquaredError)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_default_compute(fixture_metric):
    """That default scoring."""
    metric = fixture_metric
    predictions = [2.5, 0.0, 2, 8]
    references = [3, -0.5, 2, 7]
    result = metric.compute(predictions=predictions, references=references)
    assert result["mean_squared_error"] == 0.375


def test_perfect_score(fixture_metric):
    """Test score on perfect predictions."""
    metric = fixture_metric
    predictions = [3, -0.5, 2, 7]
    references = [3, -0.5, 2, 7]
    result = metric.compute(predictions=predictions, references=references)
    assert result["mean_squared_error"] == 0


def test_multioutpout_default(fixture_metric):
    """Multioutput case with errors of all outputs averaged with uniform weight."""
    metric = load_metric("mean_squared_error", config_name="multioutput")
    predictions = [[0, 2], [-1, 2], [8, -5]]
    references = [[0.5, 1], [-1, 1], [7, -6]]
    result = metric.compute(predictions=predictions, references=references)
    assert result
    assert pytest.approx(result["mean_squared_error"], 0.01) == 0.708


def test_multioutpout_raw_values(fixture_metric):
    """Multioutput case returning full set of errors."""
    metric = load_metric("mean_squared_error", config_name="multioutput")
    predictions = [[0, 2], [-1, 2], [8, -5]]
    references = [[0.5, 1], [-1, 1], [7, -6]]
    result = metric.compute(
        predictions=predictions,
        references=references,
        multioutput="raw_values",
    )
    assert result
    np.testing.assert_allclose([0.416667, 1], result["mean_squared_error"], rtol=1e-6)


def test_multioutpout_array_like(fixture_metric):
    """Multioutput case providing weights by which to average errors."""
    metric = load_metric("mean_squared_error", config_name="multioutput")
    predictions = [[0, 2], [-1, 2], [8, -5]]
    references = [[0.5, 1], [-1, 1], [7, -6]]
    result = metric.compute(
        predictions=predictions,
        references=references,
        multioutput=[0.3, 0.7],
    )
    assert result
    assert result["mean_squared_error"] == 0.825



# type: ignore

import numpy as np
from scipy.sparse import issparse
from scipy.special import logsumexp
from scipy.stats import entropy
from sklearn.naive_bayes import _BaseDiscreteNB
from sklearn.utils.extmath import safe_sparse_dot


class CorrectedNB(_BaseDiscreteNB):
    """
    Class to implement a Naive Bayes classifier with corrected probabilities
    based on the paper https://pubs.acs.org/doi/10.1021/jm0303195.

    Parameters
    ----------
    alpha : float or array-like of shape (n_features,), default=1.0
        Additive (Laplace/Lidstone) smoothing parameter
        (set alpha=0 for no smoothing).

    fit_prior : bool, default=True
        Whether to learn class prior probabilities or not.
        If false, a uniform prior will be used.

    class_prior : array-like of shape (n_classes,), default=None
        Prior probabilities of the classes. If specified, the priors are not
        adjusted according to the data.
    """

    def __init__(self, alpha=1.0, fit_prior=True, class_prior=None) -> None:
        super().__init__(
            alpha=alpha,
            fit_prior=fit_prior,
            class_prior=class_prior,
            force_alpha=True,
        )

    def _update_feature_log_prob(self, alpha: float = 1) -> None:
        """
        Calculates normalised Laplacian smoothed corrected log probabilties for
        each feature as shown in paper: https://pubs.acs.org/doi/10.1021/jm0303195.
        Concretely:

            P_corr = ((F_c + alpha)/(F_tot + alpha/(C_c/C_tot)))/(C_c/C_tot)

        where F_c is the feature frequency of feature F for class c, alpha is
        a smoothing factor (=1 by default), F_tot is the total feature
        frequency of feature F, C_c is the class count of class c, C_tot is the
        total number of samples (sum of class counts for all classes).

        Parameters
        ----------
        alpha : float = 1, optional
            The smoothing parameter to use, default is 1. Called internally
            during fitting with self.alpha.
        """
        total_samples = self.class_count_.sum()
        numerator = self.feature_count_ + alpha
        pbases = self.class_count_ / total_samples
        denominator = self.feature_count_.sum(axis=0) + (
            alpha / (pbases).reshape(-1, 1)
        )
        p_smooth = numerator / denominator
        self.feature_log_prob_ = np.log(p_smooth / pbases.reshape(-1, 1))

    @property
    def feature_entropy_(self) -> np.ndarray:
        """
        Feature entropy property. Obtained by taking the entropy of
        each features log probability for each class.

        Returns
        -------
        np.ndarray :
            The entropy of each feature, given by the feature log
            probability for each class.
        """
        return entropy(self.feature_prob_, axis=0)

    @property
    def presence_feature_entropy(self) -> np.ndarray:
        """
        Returns the entropy component for the presence of each feature in the
        dataset.
        """
        probs = np.sum(self.feature_count_, axis=1) / np.sum(self.class_count_)
        return probs * np.log(probs)

    @property
    def absence_feature_entropy(self) -> np.ndarray:
        """
        Returns the entropy component for the absence of each feature in the
        dataset.
        """
        probs = np.sum(self.class_count_) - np.sum(
            self.feature_count_,
            axis=1,
        ) / np.sum(self.class_count_)
        return probs * np.log(probs)

    @property
    def feature_prob_(self) -> np.ndarray:
        """
        Property to obtain feature probability. Exponential of
        feature log prob.

        Returns
        -------
        np.ndarray :
            The probability of the features for each class.
        """
        return np.exp(self.feature_log_prob_)

    def log_likelihood(self, X: np.ndarray) -> np.ndarray:
        """
        Returns the log likelihood for each sample in X for each class.
        Concretely, takes the dot product between X and feature log prob.

        Parameters
        ----------
        X : np.ndarray
            The binary feature data to obtain the log likelihoods
            for.

        Returns
        -------
        np.ndarray :
            The log likelihoods for each sample in X for each class.
        """
        return safe_sparse_dot(X, self.feature_log_prob_.T)

    def _joint_log_likelihood(self, X):
        return self.joint_log_likelihood(X)

    def joint_log_likelihood(self, X: np.ndarray):
        """
        Function to obtain the joint log likelihood. Equivalent to
        log(P(class)*product(P(feature_i|class))))
        Parameters
        ----------
        X : np.ndarray
            The data to obtain the joint log likelihood from.

        Returns
        -------
        np.ndarray :
            The joint log likelihood for each sample in X for each
            class.
        """
        jll = self.log_likelihood(X)
        # jll += self.class_log_prior_
        return jll

    def predict_log_proba(self, X: np.ndarray):
        """
        Function to get the log probability of each class for each
        sample in X.

        Parameters
        ----------
        X : np.ndarray
            The X data to obtain the log probabilities for.

        Returns
        -------
        np.ndarray :
            An array of shape (n_samples, n_classes) for each sample in
            X. Each sample has the log probabilities for each class.
        """
        jll = self.joint_log_likelihood(X)
        log_prob_x = logsumexp(jll, axis=1)
        return jll - np.atleast_2d(log_prob_x).T

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        """
        Same as predict_log_proba but simply the exponential of the
        values.

        Parameters
        ----------
        X : np.ndarray
            The X data to obtain the probabilities for.

        Returns
        -------
        np.ndarray :
            An array of shape (n_samples, n_classes) for each sample in
            X. Each sample has the probabilities for each class.
        """
        return np.exp(self.predict_log_proba(X))

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Function to predict labels for each sample in X. Uses the MAE
        decision rule, i.e. class with highest probability -> prediction.

        Parameters
        ----------
        X : np.ndarray
            X data to classify, must be binary.

        Returns
        -------
        np.ndarray :
            The labels from .classes_ for each sample in X.
        """
        return self.classes_[self.predict_proba(X).argmax(axis=1)]

    def entropy(self, X: np.ndarray) -> np.ndarray:
        """
        The entropy of each sample in X. Calculated by taking the
        entropy of each class probability for each sample in X.

        Parameters
        ----------
        X : np.ndarray
            The samples to obtain the entropy for. Must be binary.

        Returns
        -------
        np.ndarray :
            The entropy for each sample in X.
        """
        entropy_vals = entropy(self.predict_proba(X), axis=1)
        return entropy_vals

    def empirical_log_likelihood(self, type_="fi|c") -> np.ndarray:
        """
        Obtains the empirical log likelihoods.

        Parameters
        ----------
        type_ : str, optional {'fi|c', 'c|fi'}
            String denoting the likelihood probabilities returned.
            Options are 'fi|c' for probability of feature i given class
            c, or 'c|fi' for probability of class c given feature i.

        Returns
        -------
        np.ndarray :
            The log likelihood for either p(fi|c) or p(c|fi).
        """
        if type_ == "fi|c":
            llh = self.feature_count_ / self.class_count_.reshape(-1, 1)
        elif type_ == "c|fi":
            llh = self.feature_count_ / self.feature_count_.sum(axis=0)
        else:
            raise NotImplementedError("Likelihood type must be 'fi|c' or 'c|fi'.")
        return np.log(llh)

    def empirical_feature_entropy(self, type_: str = "fi|c") -> np.ndarray:
        """
        Calculates the empirical feature entropy by a given conditional `type_`.

        Parameters
        ----------
        type_ : str = 'fi|c', {'fi|c', 'c|fi'}
            The conditional type for the feature entropy, either feature i
            given the class `c` or class c given feature i.

        Returns
        -------
        np.ndarray :
            The empirical feature entropy.
        """
        lh = np.exp(self.empirical_log_likelihood(type_=type_))
        return entropy(lh, axis=0)

    def empirical_feature_entropy_neg_class(self, class_: int = 1, type_: str = "fi|c"):
        """
        Calculates the empirical feature entropy for a class compared to against
        the negative of the class.

        Parameters
        ----------
        class_ : int = 1, optional
            The class to use for the feature negative entropy. Default is `1`,
            i.e. class self.classes_[class_] will be used.
        type_ : str = 'fi|c', {'fi|c', 'c|fi'}
            The conditional type for the feature entropy, either feature i
            given the class `c` or class c given feature i.

        Returns
        -------
        np.ndarray :
            The empirical feature entropy for a class against 1 - the probability
            of that class.
        """
        lh = np.exp(self.empirical_log_likelihood(type_=type_))[class_]
        pk = np.array([lh, 1 - lh])
        return entropy(pk, axis=0)

    def _count(self, X, y):
        if np.any((X.data if issparse(X) else X) < 0):
            raise ValueError("Input X must be non-negative")
        self.feature_count_ += safe_sparse_dot(y.T, X)
        self.class_count_ += np.sum(y, axis=0)


class CoverageNB(CorrectedNB):
    """
    Classifier similar to `CorrectedNB` but with additional methods relevant
    to Coverage Score algorithm - such as feature coverage and feature final
    coverage.
    """

    @property
    def feature_coverage_(self) -> np.ndarray:
        """
        Property indicating the feature 'coverage'. This is calculated
        as -(ln(P(C_c|feature_i) - ln(P(C_c))) for each class c.

        Returns
        -------
        np.ndarray :
            The feature 'coverage' for each class k for each feature.
        """
        cov = -self.feature_log_prob_
        return cov

    @property
    def feature_coverage_entropy_(self) -> np.ndarray:
        """
        Returns the feature coverage multiplied by the feature entropy.

        Returns
        -------
        np.ndarray :
            The feature coverage multiplied by the feature entropy.
        """
        raw_cov = self.feature_coverage_
        shan_ents = self.get_feature_entropy_class(norm=True)
        return raw_cov * shan_ents

    @property
    def feature_final_coverage_(self) -> np.ndarray:
        """
        Property indicating the 'final coverage' of a feature.

        Returns
        -------
        np.ndarray :
            The final coverage for each feature for each class.
        """
        # Get raw coverage
        raw_cov = self.feature_coverage_
        shan_ents = self.get_feature_entropy_class(norm=True)
        # Coverage is base_cov * entropy.
        final_cov = self.feature_coverage_entropy_
        frac_count = self.feature_count_ / self.class_count_.reshape(-1, 1)
        # For frequencies > half the samples, alter coverage score (2 - entropy)
        final_cov[(frac_count > 0.5) & (raw_cov < 0)] = raw_cov[
            (frac_count > 0.5) & (raw_cov < 0)
        ] * (2 - shan_ents[(frac_count > 0.5) & (raw_cov < 0)])
        return final_cov

    def get_feature_entropy_class(
        self,
        norm: bool = True,
    ) -> np.ndarray:
        """
        Function to obtain the entropy for each feature for each class.
        Calculated as H(F_c) = -(P(C_c)*log(P(C_c)) + (1-P(C_c))*log((1-P(C_c)))
        for each class c.

        Parameters
        ----------
        norm : bool, optional
            Determines whether to normalize the entropies such that
            they fall between 0 and 1. Default is True.

        Returns
        -------
        np.ndarray :
            Returns the entropy for each feature, for each class c.
        """
        probs = np.exp(self.empirical_log_likelihood(type_="fi|c"))
        denom = np.log(2) if norm else 1
        neg_probs = 1 - probs

        return entropy([probs, neg_probs], axis=0) / denom


class PipelinePilotNB(CorrectedNB):
    """
    Naive Bayes classifier explicitly used in PipelinePilot. Main difference
    is the determination of the base probabilities for calculating a smoothed
    normalised probability.
    """

    def _update_feature_log_prob(self, alpha: float | None = None) -> None:
        """
        Updating the feature log probability to use a normalised, Laplacian
        corrected feature probability as used by PipelinePilot.
        Concretely:

            P_corr = ((F_c + alpha)/(F_tot + alpha/(FC_c/FC_tot)))/(FC_c/FC_tot)

        where F_c is the feature frequency of feature F for class c, alpha is
        a smoothing factor (=1 by default), F_tot is the total feature
        frequency of feature F, FC_c is the **total feature** count of class c,
        FC_tot is the **total feature count of all samples**
        (sum of feature counts for all classes).

        Parameters
        ----------
        alpha : float = 1, optional
            The smoothing parameter to use, default is 1. Called internally
            during fitting with self.alpha.
        """
        total_feature_counts = self.feature_count_.sum()
        summed_feature_counts = self.feature_count_.sum(axis=1)
        feature_count_tot = self.feature_count_.sum(axis=0)

        pbases = summed_feature_counts / total_feature_counts

        numerator = self.feature_count_ + alpha
        denominator = feature_count_tot + 1 / pbases.reshape(-1, 1)

        p_smooth = numerator / denominator

        self.feature_log_prob_ = np.log(p_smooth / pbases.reshape(-1, 1))



"""Uncertainty measure: Computes coefficient of variation."""

import logging
from typing import Any

import evaluate
import numpy as np
import scipy

import datasets
from molflux.metrics.bases import UncertaintyMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_CITATION = """\
@article{busk2021calibrated,
  title={Calibrated uncertainty for molecular property prediction using ensembles of message passing neural networks},
  author={Busk, Jonas and J{\\o}rgensen, Peter Bj{\\o}rn and Bhowmik, Arghya and Schmidt, Mikkel N and Winther, Ole and Vegge, Tejs},
  journal={Machine Learning: Science and Technology},
  volume={3},
  number={1},
  pages={015012},
  year={2021},
  publisher={IOP Publishing}
}
"""

_DESCRIPTION = """\
Computes coefficient of variation for a prediction interval across a dataset.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated targets as returned by a classifier/regressor.
    references: Ground truth (correct) target values.
    prediction_intervals: Prediction intervals of lower and upper bounds.

Returns:
    coefficient_of_variation: coefficient of variation (CV) indicating the heteroscedasticity or homoscedasticity of predictions across a dataset.
    High CV indicates large dispersion (high dependence of uncertainty estimates), whereas low CV indicates constant, uninformative uncertainty estimates.

Examples:
    >>> from molflux.metrics import load_metric
    >>> m = load_metric("coefficient_of_variation")
    >>> ref = [0, 1, 2, 3, 4]
    >>> pred = [0, 2, 2, 3, 5]
    >>> lower_bound = [0.1, 0.1, 0.2, 2, 3]
    >>> upper_bound = [1, 1.5, 1.8, 2.8, 5.7]
    >>> prediction_intervals = list(zip(lower_bound, upper_bound))
    >>> m.compute(references=ref, predictions=pred, prediction_intervals=prediction_intervals)
    {'coefficient_of_variation': 0.459...}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class CoefficientOfVariation(UncertaintyMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        confidence: float = 0.9,
        **kwargs: Any,
    ) -> MetricResult:
        if standard_deviations is not None and prediction_intervals is not None:
            raise ValueError(
                "Both standard deviations and prediction intervals given. Please provide only one.",
            )

        if standard_deviations is not None:
            lower_bound, upper_bound = scipy.stats.norm.interval(
                confidence,
                loc=predictions,
                scale=standard_deviations,
            )
        elif prediction_intervals is not None:
            lower_bound, upper_bound = zip(*prediction_intervals, strict=False)
        else:
            raise ValueError(
                "Either standard_deviations or prediction_intervals must be supplied.",
            )

        lower_bound = np.array(lower_bound)
        upper_bound = np.array(upper_bound)
        if (upper_bound < lower_bound).any():
            raise ValueError("Please ensure upper bound is greater than lower bound.")
        # Definition in equation 10 of reference in _CITATION relates to standard deviation, sigma.
        # Making a Gaussian assumption, then sigma is related to the prediction interval width by rescaling by a factor of scipy.norm.ppf(1-alpha/2).
        # However, this rescaling factor is in both numerator and denominator of equation 10, so drops out.
        # We can therefore work with the prediction interval width directly, which should generalise better to interval predictions that don't rely on Gaussian assumptions
        sigma = upper_bound - lower_bound
        sigma_bar = np.mean(sigma)  # average prediction interval width
        if sigma_bar == 0:
            score = 0  # avoid division by zero in case of interval of width 0
        else:
            score = np.sqrt(np.mean((sigma - sigma_bar) ** 2)) / sigma_bar
        return {self.tag: score}



import os
import pathlib
from collections.abc import Mapping
from typing import Any

import molflux.datasets
import molflux.modelzoo
import molflux.splits
from datasets import Dataset, DatasetDict
from molflux.core.featurisation.metadata import save_featurisation_metadata
from molflux.core.io import save_dict_to_json
from molflux.core.typing import FoldScores, PathLike

_FEATURISED_DATASET_FILENAME = "featurised_dataset.parquet"
_SCORES_FILENAME = "scores.json"
_PARAMS_FILENAME = "model_params.json"
_SPLITTING_STRATEGY_FILENAME = "splitting_strategy.json"
_PIPELINE_CONFIG_FILENAME = "pipeline.json"


def log_params(mapping: Mapping, path: PathLike) -> str:
    """Logs arbitrary dicts of parameters to disk."""
    return save_dict_to_json(mapping, path=path)


def log_dataset(dataset: Dataset, path: PathLike) -> str:
    """Logs an arbitrary dataset to disk."""
    format = pathlib.Path(path).suffix.lstrip(".")
    molflux.datasets.save_dataset_to_store(dataset, path=str(path), format=format)  # type: ignore[arg-type]
    return str(path)


def log_dataset_dict(
    dataset_dict: DatasetDict,
    paths_dict: dict[str, str],
) -> dict[str, str]:
    """Logs an arbitrary dataset_dict to disk"""
    save_paths: dict[str, str] = {}
    for split_name, dataset in dataset_dict.items():
        save_path = log_dataset(dataset, path=paths_dict[split_name])
        save_paths[split_name] = save_path
    return save_paths


def log_featurised_dataset(featurised_dataset: Dataset, directory: PathLike) -> str:
    """Logs a featurised dataset to disk."""
    output_file = f"{directory}/{_FEATURISED_DATASET_FILENAME}"
    return log_dataset(dataset=featurised_dataset, path=output_file)


def log_featurisation_metadata(
    featurisation_metadata: dict[str, Any] | None,
    directory: PathLike,
) -> str:
    """Logs featurisation metadata to disk.

    This can be used as an independent function to save featurisation metadata
    decoupled from the model saving process.
    """
    return save_featurisation_metadata(
        featurisation_metadata=featurisation_metadata,
        path=directory,
    )


def log_fold(fold: DatasetDict, directory: PathLike) -> dict[str, str]:
    """Logs a fold to disk as a set of .parquet files."""
    save_paths: dict[str, str] = {}
    for split, dataset in fold.items():
        _SPLIT_FILENAME = f"{split}.parquet"
        path = os.path.join(directory, _SPLIT_FILENAME)
        molflux.datasets.save_dataset_to_store(dataset, path=path, format="parquet")
        save_paths[split] = path
    return save_paths


def log_predictions(predictions: DatasetDict, directory: str) -> dict[str, str]:
    """Logs fold predictions to storage as a set of standardised parquet files."""
    paths_dict = {
        split_name: os.path.join(directory, f"{split_name}_predictions.parquet")
        for split_name in predictions.keys()
    }
    return log_dataset_dict(predictions, paths_dict=paths_dict)


def log_references(references: DatasetDict, directory: str) -> dict[str, str]:
    """Logs fold references to storage as a set of standardised parquet files."""
    paths_dict = {
        split_name: os.path.join(directory, f"{split_name}_references.parquet")
        for split_name in references.keys()
    }
    return log_dataset_dict(references, paths_dict=paths_dict)


def log_inputs(inputs: DatasetDict, directory: str) -> dict[str, str]:
    """Logs a fold's model input features to storage as parquet."""
    paths_dict = {
        split_name: os.path.join(directory, f"{split_name}_inputs.parquet")
        for split_name in inputs.keys()
    }
    return log_dataset_dict(inputs, paths_dict=paths_dict)


def log_model_params(model: molflux.modelzoo.Model, directory: PathLike) -> str:
    """Logs model metadata to disk."""
    output_file = f"{directory}/{_PARAMS_FILENAME}"
    return save_dict_to_json(model.metadata, path=output_file)


def log_pipeline_config(config: Mapping, directory: PathLike) -> str:
    """Logs the pipeline config to disk."""
    output_file = f"{directory}/{_PIPELINE_CONFIG_FILENAME}"
    return save_dict_to_json(config, path=output_file)


def log_scores(scores: FoldScores, directory: PathLike) -> str:
    """Logs metrics scores to disk."""
    output_file = f"{directory}/{_SCORES_FILENAME}"
    return save_dict_to_json(scores, path=output_file)


def log_splitting_strategy(
    strategy: molflux.splits.SplittingStrategy,
    directory: PathLike,
) -> str:
    """Logs splitting strategy metadata to disk."""
    output_file = f"{directory}/{_SPLITTING_STRATEGY_FILENAME}"

    payload = {
        "name": strategy.name,
        "tag": strategy.tag,
        "metadata": strategy.metadata,
    }

    return save_dict_to_json(payload, path=output_file)



from collections.abc import Iterable

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)

try:
    from sklearn.naive_bayes import BernoulliNB as SKBernoulliNB
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn naive bayes classifier.

BernoulliNB implements the naive Bayes training and classification algorithms for data
that is distributed according to multivariate Bernoulli distributions; i.e., there may
be multiple features but each one is assumed to be a binary-valued (Bernoulli, boolean)
variable. Therefore, this class requires samples to be represented as binary-valued
feature vectors; if handed any other kind of data, a BernoulliNB instance may
binarize its input (depending on the binarize parameter).
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
alpha : float, default=1.0
    Additive (Laplace/Lidstone) smoothing parameter
    (0 for no smoothing).
binarize : float or None, default=0.0
    Threshold for binarizing (mapping to booleans) of sample features.
    If None, input is presumed to already consist of binary vectors.
fit_prior : bool, default=True
    Whether to learn class prior probabilities or not.
    If false, a uniform prior will be used.
class_prior : array-like of shape (n_classes,), default=None
    Prior probabilities of the classes. If specified, the priors are not
    adjusted according to the data.
"""


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class BernoulliNBConfig(ModelConfig):
    alpha: float = 1
    binarize: float | None = 0
    fit_prior: bool = True
    class_prior: Iterable | None = None


class BernoulliNBClassifier(
    SKLearnClassificationMixin,
    SKLearnModelBase[BernoulliNBConfig],
):
    @property
    def _config_builder(self) -> type[BernoulliNBConfig]:
        return BernoulliNBConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKBernoulliNB:
        config = self.model_config
        return SKBernoulliNB(
            alpha=config.alpha,
            binarize=config.binarize,
            fit_prior=config.fit_prior,
            class_prior=config.class_prior,
        )



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.errors import FeaturisationError
from molflux.features.representations.openeye.canonical.smiles import (
    CanonicalSmiles,
)

representation_name = "canonical_smiles"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, CanonicalSmiles)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_compute(fixture_representation):
    """That default scoring gives expected results."""
    representation = fixture_representation
    samples = ["C#C[C@]1(O)CC[C@H]2[C@@H]3CCC4=Cc5oncc5C[C@]4(C)[C@H]3CC[C@@]21C"]
    result = representation.featurise(samples)
    expected_result = [
        "C[C@]12CC[C@H]3[C@H]([C@@H]1CC[C@]2(C#C)O)CCC4=Cc5c(cno5)C[C@]34C",
    ]
    assert representation_name in result
    assert result[representation_name] == expected_result


def test_with_remove_formal_charges_and_reasonable_protomer_raises(
    fixture_representation,
):
    """That an error is raised if attempting to remove formal charges and
    setting reasonable_protomer."""
    representation = fixture_representation
    samples = ["C#C[C@]1(O)CC[C@H]2[C@@H]3CCC4=Cc5oncc5C[C@]4(C)[C@H]3CC[C@@]21C"]

    with pytest.raises(FeaturisationError):
        representation.featurise(
            samples,
            remove_formal_charges=True,
            reasonable_protomer=True,
        )


def test_with_remove_formal_charges_and_set_neutral_ph_raises(fixture_representation):
    """That a ValueError is raised if attempting to remove formal charges and
    setting a neutral pH."""
    representation = fixture_representation
    samples = ["C#C[C@]1(O)CC[C@H]2[C@@H]3CCC4=Cc5oncc5C[C@]4(C)[C@H]3CC[C@@]21C"]

    with pytest.raises(FeaturisationError):
        representation.featurise(
            samples,
            remove_formal_charges=True,
            set_neutral_ph=True,
        )


def test_with_sd_title_tag(fixture_representation):
    """That can set a custom sd_title_tag."""
    representation = fixture_representation
    samples = ["C#C[C@]1(O)CC[C@H]2[C@@H]3CCC4=Cc5oncc5C[C@]4(C)[C@H]3CC[C@@]21C"]

    result = representation.featurise(samples, sd_title_tag="pytest")
    assert result


def test_with_clear_sd_data(fixture_representation):
    """That can set a custom sd_title_tag."""
    representation = fixture_representation
    samples = ["C#C[C@]1(O)CC[C@H]2[C@@H]3CCC4=Cc5oncc5C[C@]4(C)[C@H]3CC[C@@]21C"]

    result = representation.featurise(samples, clear_sd_data=True)
    assert result


@pytest.mark.parametrize(
    ["sample", "expected_result"],
    [
        ("[Na+].CCC(=O)[O-]", "CCC(=O)[O-]"),
        ("[Fr+].C(=O)[O-]", "C(=O)[O-]"),
        ("[Na+].[Cl-]", "[Na+]"),
        ("[Cl-].[Na+]", "[Cl-]"),
        ("O=C(O)C.O=C(O)CC", "CCC(=O)O"),
    ],
)
def test_with_strip_salts(fixture_representation, sample, expected_result):
    """That can strip salts."""
    representation = fixture_representation

    result = representation.featurise(samples=[sample], strip_salts=True)
    assert representation_name in result
    assert result[representation_name] == [expected_result]


@pytest.mark.parametrize(
    ["sample", "expected_result"],
    [
        ("C", "[H]C([H])([H])[H]"),
        ("C(=O)O", "[H]C(=O)O[H]"),
        ("C(=O)[O-]", "[H]C(=O)[O-]"),
    ],
)
def test_with_explicit_h(fixture_representation, sample, expected_result):
    """That can add explicit hydrogens"""
    representation = fixture_representation

    result = representation.featurise(samples=[sample], explicit_h=True)
    assert representation_name in result
    assert result[representation_name] == [expected_result]


def test_with_reasonable_tautomer(fixture_representation):
    """That can set reasonable tautomer."""
    representation = fixture_representation
    samples = ["CC=C(O)C"]

    result = representation.featurise(samples, reasonable_tautomer=True)
    expected_result = ["CCC(=O)C"]
    assert representation_name in result
    assert result[representation_name] == expected_result


def test_with_neutral_ph(fixture_representation):
    """That can set neutral pH"""
    representation = fixture_representation
    samples = ["CN(C)C"]

    result = representation.featurise(samples, set_neutral_ph=True)
    expected_result = ["C[NH+](C)C"]
    assert representation_name in result
    assert result[representation_name] == expected_result


def test_with_reasonable_protomer(fixture_representation):
    """That can get reasonable protomer"""
    representation = fixture_representation
    samples = ["NC(Cc1ccc(O)cc1)C(O)=O"]

    result = representation.featurise(samples, reasonable_protomer=True)
    expected_result = ["c1cc(ccc1CC(C(=O)[O-])[NH3+])O"]
    assert representation_name in result
    assert result[representation_name] == expected_result


@pytest.mark.parametrize(
    ["sample", "expected_result"],
    [
        ("[O-][H]", "O"),
        ("CC(=O)[O-]", "CC(=O)O"),
        ("C#[N-]", "C#N"),
        ("[NH+]", "[N]"),
        ("C[N+](C)(C)C", "C[N+](C)(C)C"),
    ],
)
def test_with_remove_formal_charges(fixture_representation, sample, expected_result):
    """That can remove formal charges"""
    representation = fixture_representation

    result = representation.featurise(samples=[sample], remove_formal_charges=True)
    assert representation_name in result
    assert result[representation_name] == [expected_result]


@pytest.mark.parametrize(
    ["sample", "expected_result"],
    [
        ("c1cc(ccc1C[C@@H](C(=O)O)N)O", "c1cc(ccc1CC(C(=O)O)N)O"),
        ("c1cc(ccc1C[C@H](C(=O)O)N)O", "c1cc(ccc1CC(C(=O)O)N)O"),
        ("F/C=C/F", "C(=CF)F"),
        (r"F/C=C\F", "C(=CF)F"),
    ],
)
def test_with_remove_stereo(fixture_representation, sample, expected_result):
    """That can remove formal charges"""
    representation = fixture_representation

    result = representation.featurise(samples=[sample], remove_stereo=True)
    assert representation_name in result
    assert result[representation_name] == [expected_result]


@pytest.mark.parametrize(
    ["sample", "expected_result"],
    [
        ("[C@@H]", "[CH]"),
        ("[C@H]", "[CH]"),
        ("F/C(F)=C/F", "C(=C(F)F)F"),
        ("F/C(F)=C([C@H])/F", "[CH]C(=C(F)F)F"),
    ],
)
def test_with_clear_non_chiral_stereo(fixture_representation, sample, expected_result):
    """That can remove formal charges"""
    representation = fixture_representation

    result = representation.featurise(samples=[sample], clear_non_chiral_stereo=True)
    assert representation_name in result
    assert result[representation_name] == [expected_result]


def test_with_remove_non_standard_stereo(fixture_representation):
    """That can remove non standard stereo"""
    representation = fixture_representation
    samples = ["CC[N@@H+](C)F"]

    result = representation.featurise(samples, remove_non_standard_stereo=True)
    expected_result = ["CC[NH+](C)F"]
    assert representation_name in result
    assert result[representation_name] == expected_result


def test_with_rekekulise(fixture_representation):
    """That can rekekulise"""
    representation = fixture_representation
    samples = ["c1cnccn1"]

    result = representation.featurise(samples, rekekulise=True)
    expected_result = ["C1=CN=CC=N1"]
    assert representation_name in result
    assert result[representation_name] == expected_result


def test_combine_flags_with_output_flavors(fixture_representation):
    """That can both rekekulize and assign explicit hydrogen atoms"""
    representation = fixture_representation
    samples = ["c1cnccn1"]

    result = representation.featurise(samples, explicit_h=True, rekekulise=True)
    expected_result = ["[H]C1=C(N=C(C(=N1)[H])[H])[H]"]
    assert representation_name in result
    assert result[representation_name] == expected_result


@pytest.mark.parametrize(
    ["sample", "expected_output_sample", "tautomer_options"],
    [
        ("C/C=C(/O)F", "C/C=C(/O)\\F", {"save_stereo": True}),
        ("C/C=C(/O)F", "CCC(=O)F", {}),
    ],
)
def test_with_tautomer_options(
    fixture_representation,
    sample,
    expected_output_sample,
    tautomer_options,
):
    """That tautomer options are correctly applied"""
    representation = fixture_representation
    samples = [sample]
    expected_result = [expected_output_sample]

    result = representation.featurise(
        samples,
        reasonable_tautomer=True,
        tautomer_options=tautomer_options,
    )

    assert representation_name in result
    assert result[representation_name] == expected_result


@pytest.mark.parametrize(
    ["expected_result", "tautomer_timeouts"],
    [
        (
            "C[C@@]12CCC[C@@]([C@H]1CC[C@@]34C2CC([C@@H](C3)C(=C)C4)O[C@@H]5[C@H](C([C@H]([C@@H](O5)CO)O)O[C@@H]6[C@H]([C@@H]([C@H]([C@@H](O6)CO)O)O)O)O[C@@H]7[C@H]([C@@H]([C@H]([C@@H](O7)CO)O)O)O)(C)C(=O)OC8[C@H]([C@@H]([C@H]([C@@H](O8)CO)O)O)O",
            [0.01],
        ),
        (
            "C[C@@]12CCC[C@@]([C@H]1CC[C@@]34C2CC([C@@H](C3)C(=C)C4)O[C@@H]5[C@H](C([C@H]([C@@H](O5)CO)O)O[C@@H]6[C@H]([C@@H]([C@H]([C@@H](O6)CO)O)O)O)O[C@@H]7[C@H]([C@@H]([C@H]([C@@H](O7)CO)O)O)O)(C)C(=O)OC8[C@H]([C@@H]([C@H]([C@@H](O8)CO)O)O)O",
            [10],
        ),
    ],
)
def test_with_tautomer_timeouts(
    fixture_representation,
    expected_result,
    tautomer_timeouts,
):
    """That tautomer timeouts are correctly applied"""
    representation = fixture_representation
    samples = [
        "C=C1C[C@]23CC[C@H]4[C@@](C)(CCC[C@@]4(C)C(=O)OC4O[C@@H](CO)[C@H](O)[C@@H](O)[C@@H]4O)C2CC(O[C@H]2O[C@@H](CO)[C@H](O)C(O[C@H]4O[C@@H](CO)[C@H](O)[C@@H](O)[C@@H]4O)[C@@H]2O[C@H]2O[C@@H](CO)[C@H](O)[C@@H](O)[C@@H]2O)[C@H]1C3",
    ]
    expected_result = [expected_result]

    result = representation.featurise(
        samples,
        reasonable_tautomer=True,
        tautomer_timeouts=tautomer_timeouts,
    )

    assert representation_name in result
    assert result[representation_name] == expected_result
